"""Provides the teehr accessor extending pandas DataFrames."""
import itertools
from math import pi
import pandas as pd
import geopandas as gpd
import logging
from pathlib import Path
import pandera as pa

from bokeh.plotting import figure, save, output_file, show, ColumnDataSource
from bokeh.palettes import colorblind
import xyzservices.providers as xyz

import teehr.models.pandera_dataframe_schemas as schemas

logger = logging.getLogger(__name__)

LOCATION_TOOLS = "pan, wheel_zoom, box_zoom, reset"
TIMESERIES_TOOLS = "xwheel_zoom, box_zoom, reset"


@pd.api.extensions.register_dataframe_accessor("teehr")
class TEEHRDataFrameAccessor:
    """Extends pandas DataFrame objects with visualization methods.

    Notes
    -----
    This class contains example methods for summarizing and plotting metrics
    as well as timeseries. This requires more validation in each method to
    ensure the DataFrame has the appropriate data.

    Methods operating on metrics data should start with 'metrics' and methods
    operating on timeseries data should start with 'timeseries'.
    """

    def __init__(self, pandas_obj):
        """Initialize the class."""
        logger.info("Initializing new dataframe_accessor object...")
        if not (isinstance(pandas_obj, gpd.GeoDataFrame)):
            logger.info("Adding DataFrame to accessor object.")
            self._df = pandas_obj
            self._gdf = None
            self._validate(self=self, obj=pandas_obj)
            logger.info("Object validation successful.")
        else:
            logger.info("Adding GeoDataFrame to accessor object. ")
            self._df = None
            self._gdf = pandas_obj
            self._validate(self=self, obj=pandas_obj)
            logger.info("Object validation successful.")

    @staticmethod
    def _validate(self, obj):
        """Validate the DataFrame object."""
        if 'table_type' not in obj.attrs:
            raise AttributeError(
                "No DataFrame Attribute 'table_type' defined."
                )

        if obj.attrs['table_type'] == 'primary_timeseries':

            # validate using pandera schema
            schema = schemas.primary_timeseries_schema(type='pandas')
            try:
                schema.validate(obj)
            except pa.errors.SchemaError as exc:
                raise AttributeError(
                    f"Pandera validation failed: {exc}"
                )

            # check for data
            if obj.index.size == 0:
                raise AttributeError("DataFrame must have data.")

        elif obj.attrs['table_type'] == 'secondary_timeseries':

            # validate using pandera schema
            schema = schemas.secondary_timeseries_schema(type='pandas')
            try:
                schema.validate(obj)
            except pa.errors.SchemaError as exc:
                raise AttributeError(
                    f"Pandera validation failed: {exc}"
                )

            # check for data
            if obj.index.size == 0:
                raise AttributeError("DataFrame must have data.")

        elif obj.attrs['table_type'] == 'joined_timeseries':

            # TO-DO: add validation

            raise NotImplementedError(
                "Joined_timeseries methods must be implemented."
            )

        elif obj.attrs['table_type'] == 'locations':

            # validate using pandera schema
            schema = schemas.locations_schema(type='pandas')
            try:
                schema.validate(obj)
            except pa.errors.SchemaError as exc:
                raise AttributeError(
                    f"Pandera validation failed: {exc}"
                )

            # check for data
            if obj.index.size == 0:
                raise AttributeError("GeoDataFrame must have data.")

            # convert to gdf if given df
            if not (isinstance(obj, gpd.GeoDataFrame)):
                raise NotImplementedError(f"""
                    Locations mapping does not currently support input of
                    type = {type(obj)}. Use to_geopandas() instead.
                """)

            # convert given crs to web mercator [EPSG:3857]
            target_crs = 'EPSG:3857'
            self._gdf.to_crs(target_crs, inplace=True)

        elif obj.attrs['table_type'] == 'location_attributes':

            # validate using pandera schema
            schema = schemas.location_attributes_schema(type='pandas')
            try:
                schema.validate(obj)
            except pa.errors.SchemaError as exc:
                raise AttributeError(
                    f"Pandera validation failed: {exc}"
                )

            # check for data
            if obj.index.size == 0:
                raise AttributeError("GeoDataFrame must have data.")

            # convert to gdf if given df (not implemented)
            if not (isinstance(obj, gpd.GeoDataFrame)):
                raise NotImplementedError(f"""
                    Location attributes mapping does not currently support
                    input of type = {type(obj)}. Use to_geopandas() instead.
                """)

            # convert given crs to web mercator [EPSG:3857]
            target_crs = 'EPSG:3857'
            self._gdf.to_crs(target_crs, inplace=True)

        elif obj.attrs['table_type'] == 'location_crosswalks':

            # validate using pandera schema
            schema = schemas.location_crosswalks_schema(type='pandas')
            try:
                schema.validate(obj)
            except pa.errors.SchemaError as exc:
                raise AttributeError(
                    f"Pandera validation failed: {exc}"
                )

            # check for data
            if obj.index.size == 0:
                raise AttributeError("GeoDataFrame must have data.")

            # convert to gdf if given df (not implemented)
            if not (isinstance(obj, gpd.GeoDataFrame)):
                raise NotImplementedError(f"""
                    Location attributes mapping does not currently support
                    input of type = {type(obj)}. Use to_geopandas() instead.
                """)

            # convert given crs to web mercator [EPSG:3857]
            target_crs = 'EPSG:3857'
            self._gdf.to_crs(target_crs, inplace=True)

        elif obj.attrs['table_type'] == 'metrics':

            # TO-DO: add validation

            raise NotImplementedError(
                "Metrics methods must be implemented."
            )

        else:
            table_type_str = obj.attrs['table_type']
            raise AttributeError(f"""
                Invalid table type:{table_type_str}. Visualization not
                supported.
            """)

    def _validate_path(self, output_dir):
        """Validate the output directory path."""
        logger.info("Validating output directory path.")
        if not isinstance(output_dir, Path):
            logger.info(f"""
                        Output directory must be a pathlib.Path object.
                        Path was provided as type: {type(output_dir)}.
                        Attempting to convert to pathlib.Path object.
            """)
            try:
                output_dir = Path(output_dir)
                logger.info("Path conversion successful.")
            except TypeError:
                logger.error("Path conversion failed.")

        # check for output location
        if output_dir is not None:
            if output_dir.exists():
                logger.info("Specified save directory is valid.")
            else:
                logger.info(""""
                    Specified directory does not exist.
                    Creating new directory to store figure.
                """)
                try:
                    Path(output_dir).mkdir(parents=True, exist_ok=True)
                except ValueError:
                    logger.error("Directory creation failed.")

        return output_dir

    def _timeseries_unique_values(
        self,
        variable_df: pd.DataFrame,
    ) -> dict:
        """Get dictionary of all unique values of each column."""
        logger.info("Retrieving unique values from DataFrame.")
        columns = variable_df.columns.to_list()
        unique_dict = {}
        for column in columns:
            unique_dict[column] = variable_df[column].unique().tolist()

        return unique_dict

    def _timeseries_schema(self) -> dict:
        """Get dictionary that defines plotting schema."""
        logger.info("Retrieving default plotting schema.")
        unique_variables = self._df['variable_name'].unique().tolist()
        raw_schema = {}
        filtered_schema = {}

        # get all unique combinations (primary_timeseries)
        if self._df.attrs['table_type'] == 'primary_timeseries':
            for variable in unique_variables:
                variable_df = self._df[self._df['variable_name'] == variable]
                unique_column_vals = self._timeseries_unique_values(
                    variable_df
                    )
                all_list = [
                    unique_column_vals['configuration_name'],
                    unique_column_vals['location_id'],
                    unique_column_vals['reference_time']
                    ]
                res = list(itertools.product(*all_list))
                raw_schema[variable] = res

        # get all unique combinations (secondary_timeseries)
        else:
            for variable in unique_variables:
                variable_df = self._df[self._df['variable_name'] == variable]
                unique_column_vals = self._timeseries_unique_values(
                    variable_df
                    )
                all_list = [
                    unique_column_vals['configuration_name'],
                    unique_column_vals['location_id'],
                    unique_column_vals['reference_time'],
                    unique_column_vals['member']
                    ]
                res = list(itertools.product(*all_list))
                raw_schema[variable] = res

        # primary timeseries filter routine (no data for combo)
        if self._df.attrs['table_type'] == 'primary_timeseries':
            for variable in unique_variables:
                valid_combos = []
                invalid_combos_count = 0
                var_df = self._df[self._df['variable_name'] == variable]
                for combo in raw_schema[variable]:
                    # reference_time is null
                    if pd.isnull(combo[2]):
                        temp = var_df[
                            (var_df['configuration_name'] == combo[0]) &
                            (var_df['location_id'] == combo[1]) &
                            (var_df['reference_time'].isnull())
                            ]
                        if not temp.empty:
                            valid_combos.append(combo)
                        else:
                            invalid_combos_count += 1
                    # reference_time is not null
                    else:
                        temp = var_df[
                            (var_df['configuration_name'] == combo[0]) &
                            (var_df['location_id'] == combo[1]) &
                            (var_df['reference_time'] == combo[2])
                            ]
                        if not temp.empty:
                            valid_combos.append(combo)
                        else:
                            invalid_combos_count += 1
                filtered_schema[variable] = valid_combos
                if invalid_combos_count > 0:
                    logger.info(f"""
                        Removed {invalid_combos_count} invalid combinations
                        from the schema.
                    """)

        # secondary timeseries filter routine (no data for combo)
        else:
            for variable in unique_variables:
                valid_combos = []
                invalid_combos_count = 0
                var_df = self._df[self._df['variable_name'] == variable]
                for combo in raw_schema[variable]:
                    # reference_time is null, member is null
                    if (pd.isnull(combo[2]) and pd.isnull(combo[3])):
                        temp = var_df[
                            (var_df['configuration_name'] == combo[0]) &
                            (var_df['location_id'] == combo[1]) &
                            (var_df['reference_time'].isnull()) &
                            (var_df['member'].isnull())
                            ]
                        if not temp.empty:
                            valid_combos.append(combo)
                        else:
                            invalid_combos_count += 1
                    # reference_time is null, member is not null
                    elif (pd.isnull(combo[2]) and not pd.isnull(combo[3])):
                        temp = var_df[
                            (var_df['configuration_name'] == combo[0]) &
                            (var_df['location_id'] == combo[1]) &
                            (var_df['reference_time'].isnull()) &
                            (var_df['member'] == combo[3])
                            ]
                        if not temp.empty:
                            valid_combos.append(combo)
                        else:
                            invalid_combos_count += 1
                    # reference_time is not null, member is null
                    elif (not pd.isnull(combo[2]) and pd.isnull(combo[3])):
                        temp = var_df[
                            (var_df['configuration_name'] == combo[0]) &
                            (var_df['location_id'] == combo[1]) &
                            (var_df['reference_time'] == combo[2]) &
                            (var_df['member'].isnull())
                            ]
                        if not temp.empty:
                            valid_combos.append(combo)
                        else:
                            invalid_combos_count += 1
                    # reference_time is not null, member is not null
                    else:
                        temp = var_df[
                            (var_df['configuration_name'] == combo[0]) &
                            (var_df['location_id'] == combo[1]) &
                            (var_df['reference_time'] == combo[2]) &
                            (var_df['member'] == combo[3])
                            ]
                        if not temp.empty:
                            valid_combos.append(combo)
                        else:
                            invalid_combos_count += 1
                filtered_schema[variable] = valid_combos
                if invalid_combos_count > 0:
                    logger.info(f"""
                        Removed {invalid_combos_count} invalid combinations
                        from the schema.
                    """)
        logger.info("Schema filtering complete.")
        return filtered_schema

    def _timeseries_format_plot(
        self,
        plot: figure,
    ) -> figure:
        """Format timeseries plot."""
        # x-axis
        plot.xaxis.major_label_orientation = pi/4
        plot.xaxis.axis_label_text_font_style = 'bold'

        # y-axis
        plot.yaxis.axis_label_text_font_style = 'bold'

        # title
        plot.title.text_font_size = '12pt'

        # legend
        plot.legend.location = 'top_right'
        plot.legend.border_line_width = 1
        plot.legend.border_line_color = 'black'
        plot.legend.border_line_alpha = 1.0
        plot.legend.background_fill_color = 'white'
        plot.legend.background_fill_alpha = 1.0
        plot.legend.click_policy = 'hide'

        # tools
        plot.sizing_mode = 'stretch_width'
        plot.toolbar.autohide = True

        return plot

    def _timeseries_generate_plot(
        self,
        schema: dict,
        df: pd.DataFrame,
        variable: str,
        output_dir: None,
    ) -> figure:
        """Generate a single timeseries plot."""
        logger.info("Generating timeseries plot.")

        # generate plot
        unique_units = df['unit_name'].unique().tolist()
        palette = itertools.cycle(colorblind['Colorblind'][8])
        p = figure(
            title="Click legend entry to toggle display of timeseries",
            y_axis_label=f"{variable} [{unique_units[0]}]",
            x_axis_label="Datetime",
            x_axis_type='datetime',
            tools=TIMESERIES_TOOLS,
            height=800
            )

        # add data to plot (primary timeseries)
        if self._df.attrs['table_type'] == 'primary_timeseries':
            for combo in schema[variable]:
                # reference_time is null
                if pd.isnull(combo[2]):
                    logger.info(f"Processing combination: {combo}")
                    logger.info(f"""
                                reference_time == NaT, ignoring reference_time
                                for combo: {combo}
                                """)
                    temp = df[
                        (df['configuration_name'] == combo[0]) &
                        (df['location_id'] == combo[1]) &
                        (df['reference_time'].isnull())
                        ]
                    if not temp.empty:
                        logger.info(f"Plotting data for combination: {combo}")
                        p.line(
                            temp.value_time,
                            temp.value,
                            legend_label=f"{combo[0]} - {combo[1]}",
                            line_width=1,
                            color=next(palette)
                        )
                    else:
                        logger.warning(f"No data for combination: {combo}")
                # reference_time is not null
                else:
                    logger.info(f"Processing combination: {combo}")
                    temp = df[
                        (df['configuration_name'] == combo[0]) &
                        (df['location_id'] == combo[1]) &
                        (df['reference_time'] == combo[2])
                        ]
                    if not temp.empty:
                        logger.info(f"Plotting data for combination: {combo}")
                        label = f"{combo[0]} - {combo[1]} - {combo[2]}"
                        p.line(
                            temp.value_time,
                            temp.value,
                            legend_label=f"{label}",
                            line_width=1,
                            color=next(palette)
                        )
                    else:
                        logger.warning(f"No data for combination: {combo}")

        # add data to plot (secondary timeseries)
        else:
            for combo in schema[variable]:
                # reference_time is null, member is null
                if (pd.isnull(combo[2]) and pd.isnull(combo[3])):
                    logger.info(f"Processing combination: {combo}")
                    logger.info(f"""
                                reference_time == NaT and member == None,
                                ignoring reference_time and member for
                                combo: {combo}
                                """)
                    temp = df[
                        (df['configuration_name'] == combo[0]) &
                        (df['location_id'] == combo[1]) &
                        (df['reference_time'].isnull()) &
                        (df['member'].isnull())
                        ]
                    if not temp.empty:
                        logger.info(f"Plotting data for combination: {combo}")
                        p.line(
                            temp.value_time,
                            temp.value,
                            legend_label=f"{combo[0]} - {combo[1]}",
                            line_width=1,
                            color=next(palette)
                        )
                    else:
                        logger.warning(f"No data for combination: {combo}")
                # reference_time is null, member is not null
                elif (pd.isnull(combo[2]) and not pd.isnull(combo[3])):
                    logger.info(f"Processing combination: {combo}")
                    logger.info(f"""
                                reference_time == NaT, ignoring reference_time
                                for combo: {combo}
                                """)
                    temp = df[
                        (df['configuration_name'] == combo[0]) &
                        (df['location_id'] == combo[1]) &
                        (df['reference_time'].isnull()) &
                        (df['member'] == combo[3])
                        ]
                    if not temp.empty:
                        logger.info(f"Plotting data for combination: {combo}")
                        label = f"{combo[0]} - {combo[1]} - {combo[3]}"
                        p.line(
                            temp.value_time,
                            temp.value,
                            legend_label=f"{label}",
                            line_width=1,
                            color=next(palette)
                        )
                    else:
                        logger.warning(f"No data for combination: {combo}")
                # reference_time is not null, member is null
                elif (not pd.isnull(combo[2]) and pd.isnull(combo[3])):
                    logger.info(f"Processing combination: {combo}")
                    logger.info(f"""
                                member == None, ignoring member for
                                combo: {combo}
                                """)
                    temp = df[
                        (df['configuration_name'] == combo[0]) &
                        (df['location_id'] == combo[1]) &
                        (df['reference_time'] == combo[2]) &
                        (df['member'].isnull())
                        ]
                    if not temp.empty:
                        logger.info(f"Plotting data for combination: {combo}")
                        label = f"{combo[0]} - {combo[1]} - {combo[2]}"
                        p.line(
                            temp.value_time,
                            temp.value,
                            legend_label=f"{label}",
                            line_width=1,
                            color=next(palette)
                        )
                    else:
                        logger.warning(f"No data for combination: {combo}")
                # reference_time is not null, member is not null
                else:
                    logger.info(f"Processing combination: {combo}")
                    temp = df[
                        (df['configuration_name'] == combo[0]) &
                        (df['location_id'] == combo[1]) &
                        (df['reference_time'] == combo[2]) &
                        (df['member'] == combo[3])
                        ]
                    if not temp.empty:
                        logger.info(f"Plotting data for combination: {combo}")
                        label = f"{combo[0]} - {combo[1]} - {combo[2]} - \
                                  {combo[3]}"
                        p.line(
                            temp.value_time,
                            temp.value,
                            legend_label=f"{label}",
                            line_width=1,
                            color=next(palette)
                        )
                    else:
                        logger.warning(f"No data for combination: {combo}")

        # format plot
        p = self._timeseries_format_plot(plot=p)

        # output figure (primary timeseries)
        if self._df.attrs['table_type'] == 'primary_timeseries':
            if output_dir:
                fname = Path(
                    output_dir,
                    f'primary_timeseries_plot_{variable}.html'
                    )
                output_file(
                    filename=fname,
                    title=f'Primary Timeseries Plot [{variable}]',
                    mode='inline'
                        )
                logger.info(f"Saving primary timeseries plot at {output_dir}")
                save(p)
                logger.info(f"Primary Timeseries plot saved at {fname}")
            else:
                logger.info("No output directory specified, displaying plot.")
                show(p)
                logger.info("Primary Timeseries plot displayed.")

        # output figure (secondary timeseries)
        else:
            if output_dir:
                fname = Path(
                    output_dir,
                    f'secondary_timeseries_plot_{variable}.html'
                    )
                output_file(
                    filename=fname,
                    title=f'Secondary Timeseries Plot [{variable}]',
                    mode='inline'
                        )
                logger.info(
                    f"Saving secondary timeseries plot at {output_dir}"
                    )
                save(p)
                logger.info(f"Secondary Timeseries plot saved at {fname}")
            else:
                logger.info("No output directory specified, displaying plot.")
                show(p)
                logger.info("Secondary Timeseries plot displayed.")

        return

    def timeseries_plot(self, output_dir=None):
        """
        Generate and save TS plots for each unique variable in theDataFrame.

        This method generates timeseries plots for each unique variable in the
        DataFrame's 'variable_name' column. The plots are saved to the
        specified output directory if provided. If the output directory does
        not exist, it will be created.

        Parameters
        ----------
        output_dir : pathlib.Path or None, optional
            The directory where the plots will be saved. If None, the plots
            will be displayed interactively. Default is None.

        Returns
        -------
        None

        Notes
        -----
        This method calls ``_timeseries_schema`` to get the plotting
        schema and ``_timeseries_generate_plot`` to generate each plot. It
        ensures the output directory exists before saving the plots.
        """
        # check table type
        if (self._df.attrs['table_type'] != 'primary_timeseries' and
           self._df.attrs['table_type'] != 'secondary_timeseries'):
            table_type_str = self.attrs['table_type']
            raise AttributeError(f"""
                Expected table_type == "primary_timeseries" or
                "secondary_timeseries", got table_type = {table_type_str}
            """)

        # validate output location
        output_dir = self._validate_path(output_dir)

        # generate plots
        schema = self._timeseries_schema()
        for variable in schema.keys():
            df_variable = self._df[self._df['variable_name'] == variable]
            self._timeseries_generate_plot(
                schema=schema,
                df=df_variable,
                variable=variable,
                output_dir=output_dir
                )

    def _location_format_points(self) -> dict:
        """Generate dictionary for point plotting."""
        logger.info("Assembling geodata for locations mapping...")
        geo_data = {}
        geo_data['id'] = self._gdf['id'].tolist()
        geo_data['name'] = self._gdf['name'].tolist()
        geo_data['x'] = self._gdf.geometry.x.values.tolist()
        geo_data['y'] = self._gdf.geometry.y.values.tolist()
        logger.info("Locations geodata assembled.")

        return geo_data

    def _location_get_bounds(
        self,
        geo_data: dict
    ) -> dict:
        """Determine axes ranges using point data."""
        logger.info("Retrieving axes ranges from geodata.")
        min_x = min(geo_data['x'])
        max_x = max(geo_data['x'])
        min_y = min(geo_data['y'])
        max_y = max(geo_data['y'])
        x_buffer = abs((max_x - min_x)*0.1)
        y_buffer = abs((max_y - min_y)*0.1)
        axes_bounds = {}
        axes_bounds['x_space'] = ((min_x - x_buffer), (max_x + x_buffer))
        axes_bounds['y_space'] = ((min_y - y_buffer), (max_y + y_buffer))

        return axes_bounds

    def _location_generate_map(
        self,
        geo_data: dict,
        output_dir: None
    ) -> figure:
        """Generate location map."""
        logger.info("Generating location map.")

        # set tooltips
        tooltips = [
            ("id", "@id"),
            ("name", "@name")
            ]

        # get axes bounds
        axes_bounds = self._location_get_bounds(geo_data=geo_data)

        # generate basemap
        p = figure(
            x_range=axes_bounds['x_space'],
            y_range=axes_bounds['y_space'],
            x_axis_type="mercator",
            y_axis_type="mercator",
            tooltips=tooltips,
            tools=LOCATION_TOOLS
            )
        p.add_tile(xyz.OpenStreetMap.Mapnik)

        # add data
        source = ColumnDataSource(data=geo_data)
        p.scatter(
            x='x',
            y='y',
            color='blue',
            source=source,
            size=10,
            fill_alpha=1.0
            )

        # output figure
        if output_dir:
            fname = Path(output_dir, 'location_map.html')
            output_file(filename=fname, title='Location Map', mode='inline')
            logger.info(f"Saving location map at {output_dir}")
            save(p)
            logger.info(f"Location map saved at {fname}")
        else:
            logger.info("No output directory specified, displaying plot.")
            show(p)
            logger.info("Location map displayed.")

        return

    def locations_map(self, output_dir=None):
        """
        Generate a location map and save it to the specified directory.

        Parameters
        ----------
        output_dir : str or Path, optional
            The directory where the generated map will be saved. If not
            provided, the map will not be saved. If the directory does not
            exist, it will be created.

        Raises
        ------
        AttributeError
            If the table type is not 'locations'.

        Notes
        -----
        This function checks the table type to ensure it is 'locations'. If an
        output directory is specified, it checks if the directory exists and
        creates it if it does not. The function then formats the location
        points and generates the map, saving it to the specified directory if
        provided.
        """
        # check table type
        if self._gdf.attrs['table_type'] != 'locations':
            table_type_str = self.attrs['table_type']
            raise AttributeError(f"""
                Expected table_type == "locations",
                got table_type = {table_type_str}
            """)

        # validate output location
        output_dir = self._validate_path(output_dir)

        geo_data = self._location_format_points()

        # generate map
        self._location_generate_map(geo_data=geo_data, output_dir=output_dir)

    def _location_attributes_format_points(self) -> dict:
        """Format location_attributes data for use in mapping method."""
        logger.info("Assembling geodata for location attributes mapping...")
        geo_data = {}
        locations = self._gdf['location_id'].unique()
        for location in locations:
            local_attributes = {}
            location_gdf = self._gdf[self._gdf['location_id'] == location]
            attributes = location_gdf['attribute_name'].unique()
            for attribute in attributes:
                row = location_gdf[location_gdf['attribute_name'] == attribute]
                local_attributes[attribute] = row['value'].values[0]
                if attribute == attributes[-1]:
                    local_attributes['x'] = row.geometry.x.values[0]
                    local_attributes['y'] = row.geometry.y.values[0]
                    local_attributes['location_id'] = location
            geo_data[location] = local_attributes
        logger.info("Location attributes geodata assembled.")
        logger.info("Checking for duplicate attributes...")
        self._location_attributes_check_duplicates(geo_data=geo_data)

        return geo_data

    def _location_attributes_check_duplicates(
        self,
        geo_data: dict
    ):
        """Check for duplicate attributes."""
        for location in geo_data.keys():
            location_data = self._gdf[self._gdf['location_id'] == location]
            attribute_counts = location_data['attribute_name'].value_counts()
            for i in range(len(attribute_counts)):
                attribute_val = attribute_counts.iloc[i]
                attribute_name = attribute_counts.index[i]
                if attribute_val > 1:
                    logger.warning(f"""
        Location = {location} has multiple entries for
        attribute = {attribute_name}. Remove duplicates to avoid unintended
        results.
                                   """)

    def _location_attributes_get_tooltips(
            self,
            geo_data: dict
    ) -> list:
        """Dynamically create tooltips for location_attributes."""
        # extract attributes from geo_data
        all_attributes = []
        for location in geo_data.keys():
            attributes = geo_data[location]
            for attribute in attributes.keys():
                all_attributes.append(attribute)

        # remove duplicates and x/y
        unique_attributes = list(set(all_attributes))
        unique_attributes.remove('x')
        unique_attributes.remove('y')

        # format tuples for tooltips
        tooltips = []
        for attribute in unique_attributes:
            entry = (f"{attribute}", f"@{attribute}")
            tooltips.append(entry)

        return tooltips

    def _location_attributes_get_bounds(
            self,
            geo_data: dict
    ) -> dict:
        """Obtain axes bounds for location_attributes mapping."""
        x_list = []
        y_list = []
        for location in geo_data.keys():
            attributes = geo_data[location]
            x_list.append(attributes['x'])
            y_list.append(attributes['y'])
        min_x = min(x_list)
        max_x = max(x_list)
        min_y = min(y_list)
        max_y = max(y_list)
        x_buffer = abs((max_x - min_x)*0.1)
        y_buffer = abs((max_y - min_y)*0.1)
        axes_bounds = {}
        axes_bounds['x_space'] = ((min_x - x_buffer), (max_x + x_buffer))
        axes_bounds['y_space'] = ((min_y - y_buffer), (max_y + y_buffer))

        return axes_bounds

    def _location_attributes_make_iterable(self, geo_data) -> dict:
        """Make values iterable for plotting."""
        for key, value in geo_data.items():
            if not isinstance(value, (list, tuple, set)):
                geo_data[key] = [value]

        return geo_data

    def _location_attributes_generate_map(
            self,
            geo_data: dict,
            output_dir=None
    ) -> figure:
        """Generate map for location_attributes table."""
        logger.info("Generating location attributes map...")

        # set tooltips
        tooltips = self._location_attributes_get_tooltips(geo_data=geo_data)

        # get axes bounds
        axes_bounds = self._location_attributes_get_bounds(geo_data=geo_data)

        # generate basemap
        p = figure(
            x_range=axes_bounds['x_space'],
            y_range=axes_bounds['y_space'],
            x_axis_type="mercator",
            y_axis_type="mercator",
            tooltips=tooltips,
            tools=LOCATION_TOOLS
            )
        p.add_tile(xyz.OpenStreetMap.Mapnik)

        # add data per location
        for location in geo_data.keys():
            location_dict = self._location_attributes_make_iterable(
                geo_data[location]
                )
            source = ColumnDataSource(location_dict)
            p.scatter(
                x='x',
                y='y',
                color='blue',
                source=source,
                size=10,
                fill_alpha=1.0
            )

        # output figure
        if output_dir:
            fname = Path(output_dir, 'location_attributes_map.html')
            output_file(
                filename=fname,
                title='Location Attributes Map',
                mode='inline'
                )
            logger.info(f"Saving location attributes map at {output_dir}")
            save(p)
            logger.info(f"Location attributes map at {fname}")
        else:
            logger.info("No output directory specified, displaying plot.")
            show(p)
            logger.info("Location attributes map displayed.")

        return

    def location_attributes_map(self, output_dir=None):
        """Generate location_attributes table map.

        Generate a map of location attributes and save it to the specified
        directory.

        This function checks the table type to ensure it is
        'location_attributes'. If an output directory is specified, it checks
        if the directory exists and creates it if it does not. It then formats
        the point data and generates the map.

        Parameters
        ----------
        output_dir : Path or None, optional
            The directory where the generated map will be saved. If None, the
            map will not be saved to a file. Default is None.

        Raises
        ------
        AttributeError
            If the table type is not 'location_attributes'.

        Notes
        -----
        This function relies on the following methods:

        - ``_location_attributes_format_points``: Formats the point data.
        - ``_location_attributes_generate_map``: Generates the map using the
                                                 formatted data.

        Examples
        --------
        >>> obj = YourClass()
        >>> obj.location_attributes_map(output_dir=Path('/path/to/save'))
        """
        # check table type
        if self._gdf.attrs['table_type'] != 'location_attributes':
            table_type_str = self.attrs['table_type']
            raise AttributeError(f"""
                Expected table_type == "location_attributes",
                got table_type = {table_type_str}
            """)

        # validate output location
        output_dir = self._validate_path(output_dir)

        # format point data
        geo_data = self._location_attributes_format_points()

        # generate map
        self._location_attributes_generate_map(
            geo_data=geo_data,
            output_dir=output_dir
            )

    def _location_crosswalks_format_points(self) -> dict:
        """Generate dictionary for point plotting."""
        logger.info("Assembling geodata for location_crosswalks mapping...")
        geo_data = {}
        geo_data['primary_id'] = self._gdf['primary_location_id'].tolist()
        geo_data['secondary_id'] = self._gdf['secondary_location_id'].tolist()
        geo_data['x'] = self._gdf.geometry.x.values.tolist()
        geo_data['y'] = self._gdf.geometry.y.values.tolist()
        logger.info("location_crosswalks geodata assembled.")

        return geo_data

    def _location_crosswalks_generate_map(
            self,
            geo_data: dict,
            output_dir=None
    ) -> figure:
        """Generate map for location_crosswalks table."""
        logger.info("Generating location_crosswalks map...")

        # set tooltips
        tooltips = [
            ("primary_location_id", "@primary_id"),
            ("secondary_location_id", "@secondary_id")
            ]

        # get axes bounds
        axes_bounds = self._location_get_bounds(geo_data=geo_data)

        # generate basemap
        p = figure(
            x_range=axes_bounds['x_space'],
            y_range=axes_bounds['y_space'],
            x_axis_type="mercator",
            y_axis_type="mercator",
            tooltips=tooltips,
            tools=LOCATION_TOOLS
            )
        p.add_tile(xyz.OpenStreetMap.Mapnik)

        # add data
        source = ColumnDataSource(data=geo_data)
        p.scatter(
            x='x',
            y='y',
            color='blue',
            source=source,
            size=10,
            fill_alpha=1.0
            )

        # output figure
        if output_dir:
            fname = Path(output_dir, 'location_crosswalks_map.html')
            output_file(
                filename=fname,
                title='Location Crosswalks Map',
                mode='inline'
                )
            logger.info(f"Saving location crosswalks map at {output_dir}")
            save(p)
            logger.info(f"Location crosswalks map saved at {fname}")
        else:
            logger.info("No output directory specified, displaying plot.")
            show(p)
            logger.info("Location crosswalks map displayed.")

        return

    def location_crosswalks_map(self, output_dir=None):
        """Generate location_crosswalks table map.

        Generate a map of location crosswalks and save it to the specified
        directory.

        This method checks the table type to ensure it is
        'location_crosswalks'. If an output directory is specified, it checks
        if the directory exists and creates it if it does not. It then
        assembles the point data and generates the map.

        Parameters
        ----------
        output_dir : Path or None, optional
            The directory where the generated map will be saved. If None, the
            map will not be saved to a file. Default is None.

        Raises
        ------
        AttributeError
            If the table type is not 'location_crosswalks'.

        Notes
        -----
        This method relies on the following methods:

        - ``_location_crosswalks_format_points``: Assembles the point data.
        - ``_location_crosswalks_generate_map``: Generates the map using the
                                               assembled data.

        Examples
        --------
        >>> obj = YourClass()
        >>> obj.location_crosswalks_map(output_dir=Path('/path/to/save'))
        """
        # check table type
        if self._gdf.attrs['table_type'] != 'location_crosswalks':
            table_type_str = self.attrs['table_type']
            raise AttributeError(f"""
                Expected table_type == "location_crosswalks",
                got table_type = {table_type_str}
            """)

        # validate output location
        output_dir = self._validate_path(output_dir)

        # assemble data
        geo_data = self._location_crosswalks_format_points()

        # generate map
        self._location_crosswalks_generate_map(
            geo_data=geo_data,
            output_dir=output_dir
            )



"""Module describing NWM v2.2 point data configuration variables."""
from teehr.models.str_enum import StrEnum

from typing import Optional

from pydantic import BaseModel


# VARIABLE ENUMS: Potential variable names for each output_type
class ChannelRtVariableEnum(StrEnum):
    """ChannelRtVariableEnum."""
    nudge = "nudge"
    qBtmVertRunoff = "qBtmVertRunoff"
    qBucket = "qBucket"
    qSfcLatRunoff = "qSfcLatRunoff"
    streamflow = "streamflow"
    velocity = "velocity"


class ChannelRtNoDAVariableEnum(StrEnum):
    """ChannelRtNoDAVariableEnum."""
    nudge = "nudge"
    qBucket = "qBucket"
    qSfcLatRunoff = "qSfcLatRunoff"
    streamflow = "streamflow"
    velocity = "velocity"


class ChannelRtLongVariableEnum(StrEnum):
    """ChannelRtLongVariableEnum."""
    nudge = "nudge"
    streamflow = "streamflow"
    velocity = "velocity"


class TerrainRtVariableEnum(StrEnum):
    """TerrainRtVariableEnum."""
    sfcheadsubrt = "sfcheadsubrt"
    zwattablrt = "zwattablrt"


class ReservoirVariableEnum(StrEnum):
    """ReservoirVariableEnum."""
    inflow = "inflow"
    outflow = "outflow"
    reservoir_assimiated_value = "reservoir_assimiated_value"
    water_sfc_elev = "water_sfc_elev"


# OUTPUT ENUMS: Potential output names for each configuration
class ShortAndAnalysisOutputEnum(StrEnum):
    """ShortAndAnalysisOutputEnum."""
    channel_rt = "channel_rt"
    terrain_rt = "terrain_rt"
    reservoir = "reservoir"


class Medium1OutputEnum(StrEnum):
    """Medium1OutputEnum."""
    channel_rt_1 = "channel_rt_1"
    terrain_rt_1 = "terrain_rt_1"
    reservoir_1 = "reservoir_1"


class Medium2OutputEnum(StrEnum):
    """Medium2OutputEnum."""
    channel_rt_2 = "channel_rt_2"
    terrain_rt_2 = "terrain_rt_2"
    reservoir_2 = "reservoir_2"


class Medium3OutputEnum(StrEnum):
    """Medium3OutputEnum."""
    channel_rt_3 = "channel_rt_3"
    terrain_rt_3 = "terrain_rt_3"
    reservoir_3 = "reservoir_3"


class Medium4OutputEnum(StrEnum):
    """Medium4OutputEnum."""
    channel_rt_4 = "channel_rt_4"
    terrain_rt_4 = "terrain_rt_4"
    reservoir_4 = "reservoir_4"


class Medium5OutputEnum(StrEnum):
    """Medium5OutputEnum."""
    channel_rt_5 = "channel_rt_5"
    terrain_rt_5 = "terrain_rt_5"
    reservoir_5 = "reservoir_5"


class Medium6OutputEnum(StrEnum):
    """Medium6OutputEnum."""
    channel_rt_6 = "channel_rt_6"
    terrain_rt_6 = "terrain_rt_6"
    reservoir_6 = "reservoir_6"


class Medium7OutputEnum(StrEnum):
    """Medium7OutputEnum."""
    channel_rt_7 = "channel_rt_7"
    terrain_rt_7 = "terrain_rt_7"
    reservoir_7 = "reservoir_7"


class MediumNoDAEnum(StrEnum):
    """MediumNoDAEnum."""
    channel_rt = "channel_rt"


class Long1OutputEnum(StrEnum):
    """Long1OutputEnum."""
    channel_rt_1 = "channel_rt_1"
    reservoir_1 = "reservoir_1"


class Long2OutputEnum(StrEnum):
    """Long2OutputEnum."""
    channel_rt_2 = "channel_rt_2"
    reservoir_2 = "reservoir_2"


class Long3OutputEnum(StrEnum):
    """Long3OutputEnum."""
    channel_rt_3 = "channel_rt_3"
    reservoir_3 = "reservoir_3"


class Long4OutputEnum(StrEnum):
    """Long4OutputEnum."""
    channel_rt_4 = "channel_rt_4"
    reservoir_4 = "reservoir_4"


# POINT OUTPUT TYPE MODELS (needed for each configuration enum)
class ShortAndAnalysis(BaseModel):
    """ShortAndAnalysis."""
    output_type: ShortAndAnalysisOutputEnum
    channel_rt: Optional[ChannelRtVariableEnum] = None
    terrain_rt: Optional[TerrainRtVariableEnum] = None
    reservoir: Optional[ReservoirVariableEnum] = None


class ShortAndAnalysisNoDA(BaseModel):
    """ShortAndAnalysisNoDA."""
    output_type: ShortAndAnalysisOutputEnum
    channel_rt: Optional[ChannelRtNoDAVariableEnum] = None
    terrain_rt: Optional[TerrainRtVariableEnum] = None
    reservoir: Optional[ReservoirVariableEnum] = None


class MediumRangeMem1(BaseModel):
    """MediumRangeMem1."""
    output_type: Medium1OutputEnum
    channel_rt_1: Optional[ChannelRtVariableEnum] = None
    terrain_rt_1: Optional[TerrainRtVariableEnum] = None
    reservoir_1: Optional[ReservoirVariableEnum] = None


class MediumRangeMem2(BaseModel):
    """MediumRangeMem2."""
    output_type: Medium2OutputEnum
    channel_rt_2: Optional[ChannelRtVariableEnum] = None
    terrain_rt_2: Optional[TerrainRtVariableEnum] = None
    reservoir_2: Optional[ReservoirVariableEnum] = None


class MediumRangeMem3(BaseModel):
    """MediumRangeMem3."""
    output_type: Medium3OutputEnum
    channel_rt_3: Optional[ChannelRtVariableEnum] = None
    terrain_rt_3: Optional[TerrainRtVariableEnum] = None
    reservoir_3: Optional[ReservoirVariableEnum] = None


class MediumRangeMem4(BaseModel):
    """MediumRangeMem4."""
    output_type: Medium4OutputEnum
    channel_rt_4: Optional[ChannelRtVariableEnum] = None
    terrain_rt_4: Optional[TerrainRtVariableEnum] = None
    reservoir_4: Optional[ReservoirVariableEnum] = None


class MediumRangeMem5(BaseModel):
    """MediumRangeMem5."""
    output_type: Medium5OutputEnum
    channel_rt_5: Optional[ChannelRtVariableEnum] = None
    terrain_rt_5: Optional[TerrainRtVariableEnum] = None
    reservoir_5: Optional[ReservoirVariableEnum] = None


class MediumRangeMem6(BaseModel):
    """MediumRangeMem6."""
    output_type: Medium6OutputEnum
    channel_rt_6: Optional[ChannelRtVariableEnum] = None
    terrain_rt_6: Optional[TerrainRtVariableEnum] = None
    reservoir_6: Optional[ReservoirVariableEnum] = None


class MediumRangeMem7(BaseModel):
    """MediumRangeMem7."""
    output_type: Medium7OutputEnum
    channel_rt_7: Optional[ChannelRtVariableEnum] = None
    terrain_rt_7: Optional[TerrainRtVariableEnum] = None
    reservoir_7: Optional[ReservoirVariableEnum] = None


class MediumRangeNoDA(BaseModel):
    """MediumRangeNoDA."""
    output_type: MediumNoDAEnum
    channel_rt: Optional[ChannelRtNoDAVariableEnum] = None


class LongRangeMem1(BaseModel):
    """LongRangeMem1."""
    output_type: Long1OutputEnum
    channel_rt_1: Optional[ChannelRtLongVariableEnum] = None
    reservoir_1: Optional[ReservoirVariableEnum] = None


class LongRangeMem2(BaseModel):
    """LongRangeMem2."""
    output_type: Long2OutputEnum
    channel_rt_2: Optional[ChannelRtLongVariableEnum] = None
    reservoir_2: Optional[ReservoirVariableEnum] = None


class LongRangeMem3(BaseModel):
    """LongRangeMem3."""
    output_type: Long3OutputEnum
    channel_rt_3: Optional[ChannelRtLongVariableEnum] = None
    reservoir_3: Optional[ReservoirVariableEnum] = None


class LongRangeMem4(BaseModel):
    """LongRangeMem4."""
    output_type: Long4OutputEnum
    channel_rt_4: Optional[ChannelRtLongVariableEnum] = None
    reservoir_4: Optional[ReservoirVariableEnum] = None


# POINT CONFIGURATION ENUM: Potential configuration names
class ConfigurationsEnum(StrEnum):
    """ConfigurationsEnum."""
    analysis_assim = "analysis_assim"
    analysis_assim_no_da = "analysis_assim_no_da"
    analysis_assim_extend = "analysis_assim_extend"
    analysis_assim_extend_no_da = "analysis_assim_extend_no_da"
    analysis_assim_long = "analysis_assim_long"
    analysis_assim_long_no_da = "analysis_assim_long_no_da"
    analysis_assim_hawaii = "analysis_assim_hawaii"
    analysis_assim_hawaii_no_da = "analysis_assim_hawaii_no_da"
    analysis_assim_puertorico = "analysis_assim_puertorico"
    analysis_assim_puertorico_no_da = "analysis_assim_puertorico_no_da"
    short_range = "short_range"
    short_range_hawaii = "short_range_hawaii"
    short_range_puertorico = "short_range_puertorico"
    short_range_hawaii_no_da = "short_range_hawaii_no_da"
    short_range_puertorico_no_da = "short_range_puertorico_no_da"
    medium_range_mem1 = "medium_range_mem1"
    medium_range_mem2 = "medium_range_mem2"
    medium_range_mem3 = "medium_range_mem3"
    medium_range_mem4 = "medium_range_mem4"
    medium_range_mem5 = "medium_range_mem5"
    medium_range_mem6 = "medium_range_mem6"
    medium_range_mem7 = "medium_range_mem7"
    medium_range_no_da = "medium_range_no_da"
    long_range_mem1 = "long_range_mem1"
    long_range_mem2 = "long_range_mem2"
    long_range_mem3 = "long_range_mem3"
    long_range_mem4 = "long_range_mem4"


# POINT CONFIGURATION MODEL
class PointConfigurationModel(BaseModel):
    """NWM v2.2 PointConfigurationModel."""
    configuration: ConfigurationsEnum = None
    analysis_assim: Optional[ShortAndAnalysis] = None
    analysis_assim_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_extend: Optional[ShortAndAnalysis] = None
    analysis_assim_extend_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_long: Optional[ShortAndAnalysis] = None
    analysis_assim_long_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_hawaii: Optional[ShortAndAnalysis] = None
    analysis_assim_hawaii_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_puertorico: Optional[ShortAndAnalysis] = None
    analysis_assim_puertorico_no_da: Optional[ShortAndAnalysisNoDA] = None
    short_range: Optional[ShortAndAnalysis] = None
    short_range_hawaii: Optional[ShortAndAnalysis] = None
    short_range_puertorico: Optional[ShortAndAnalysis] = None
    short_range_hawaii_no_da: Optional[ShortAndAnalysisNoDA] = None
    short_range_puertorico_no_da: Optional[ShortAndAnalysisNoDA] = None
    medium_range_mem1: Optional[MediumRangeMem1] = None
    medium_range_mem2: Optional[MediumRangeMem2] = None
    medium_range_mem3: Optional[MediumRangeMem3] = None
    medium_range_mem4: Optional[MediumRangeMem4] = None
    medium_range_mem5: Optional[MediumRangeMem5] = None
    medium_range_mem6: Optional[MediumRangeMem6] = None
    medium_range_mem7: Optional[MediumRangeMem7] = None
    medium_range_no_da: Optional[MediumRangeNoDA] = None
    long_range_mem1: Optional[LongRangeMem1] = None
    long_range_mem2: Optional[LongRangeMem2] = None
    long_range_mem3: Optional[LongRangeMem3] = None
    long_range_mem4: Optional[LongRangeMem4] = None


if __name__ == "__main__":
    # So for example:
    configuration = "medium_range_mem1"
    output_type = "channel_rt_1"
    variable_name = "streamflow"

    # Assemble input parameters
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }

    # Check input parameters
    cm = PointConfigurationModel.model_validate(vars)

    config = cm.configuration.name
    forecast_obj = getattr(cm, config)
    out_type = forecast_obj.output_type.name
    var_name = getattr(forecast_obj, out_type).name

    pass



"""Enums and Basemodels for metric classes."""
from typing import Union, Callable

from teehr.models.str_enum import StrEnum
from teehr.querying.utils import unpack_sdf_dict_columns

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, ConfigDict, model_validator
from pyspark.sql import types as T


# Pydantic BaseModel configurations
class MetricsBasemodel(PydanticBaseModel):
    """Metrics Basemodel configuration."""

    return_type: Union[str, T.ArrayType, T.MapType] = Field(default=None)
    unpack_results: bool = Field(default=False)
    unpack_function: Callable = Field(default=None)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra='forbid'  # raise an error if extra fields are passed
    )


class ProbabilisticBasemodel(MetricsBasemodel):
    """Probabilistic Basemodel configuration."""

    @model_validator(mode="before")
    def update_return_type(cls, values):
        """Update the return type based on the summary function."""
        if values.get("summary_func") is None:
            values["return_type"] = T.ArrayType(T.FloatType())
        elif values.get("summary_func") is not None:
            values["return_type"] = "float"
        return values


class BootstrapBasemodel(MetricsBasemodel):
    """Bootstrap Basemodel configuration."""

    @model_validator(mode="before")
    def update_return_type(cls, values):
        """Update the return type based on the quantiles."""
        if values.get("quantiles") is None:
            values["return_type"] = T.ArrayType(T.FloatType())
        elif values.get("quantiles") is not None:
            values["return_type"] = T.MapType(T.StringType(), T.FloatType())
        return values


class DeterministicBasemodel(MetricsBasemodel):
    """Deterministic Basemodel configuration."""

    unpack_function: Callable = Field(default=unpack_sdf_dict_columns)
    return_type:  Union[str, T.ArrayType, T.MapType] = Field(default="float")


# Enums
class Operators(StrEnum):
    """Operators for filters."""

    eq = "="
    gt = ">"
    lt = "<"
    gte = ">="
    lte = "<="
    islike = "like"
    isin = "in"


class ValueTypeEnum(StrEnum):
    """Value types."""

    Deterministic = "Deterministic"
    Categorical = "Categorical"
    Signature = "Signature"
    Probabilistic = "Probabilistic"
    Time = "Time"


class BootstrapMethodEnum(StrEnum):
    """Bootstrap methods."""

    percentile = "percentile"
    bias_corrected = "bias_corrected"
    bias_corrected_percentile = "bias_corrected_percentile"
    t = "t"
    pivot = "pivot"


class TransformEnum(StrEnum):
    """Transform methods."""

    log = "log"
    sqrt = "sqrt"
    square = "square"
    cube = "cube"
    exp = "exp"
    inv = "inv"
    abs = "abs"
    none = "none"


class CRPSEstimators(StrEnum):
    """CRPS Estimators."""

    pwm = "pwm"
    nrg = "nrg"
    fair = "fair"


class MetricCategories(StrEnum):
    """Metric categories."""

    Deterministic = "Deterministic"
    Categorical = "Categorical"
    Signature = "Signature"
    Probabilistic = "Probabilistic"




import teehr.const as const
from teehr.evaluation.tables.base_table import BaseTable
from teehr.loading.location_crosswalks import convert_location_crosswalks
from teehr.loading.utils import validate_input_is_csv, validate_input_is_parquet
from teehr.models.filters import LocationCrosswalkFilter
from teehr.models.table_enums import LocationCrosswalkFields
from teehr.querying.utils import join_geometry
import teehr.models.pandera_dataframe_schemas as schemas
from pathlib import Path
from typing import Union
import logging
from teehr.utils.utils import to_path_or_s3path, remove_dir_if_exists
from teehr.models.table_enums import TableWriteEnum
from teehr.loading.utils import add_or_replace_sdf_column_prefix


logger = logging.getLogger(__name__)


class LocationCrosswalkTable(BaseTable):
    """Access methods to location crosswalks table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "location_crosswalks"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.format = "parquet"
        self.filter_model = LocationCrosswalkFilter
        self.schema_func = schemas.location_crosswalks_schema
        self.unique_column_set = [
            "primary_location_id",
            "secondary_location_id"
        ]

    def _load(
        self,
        in_path: Union[Path, str],
        field_mapping: dict = None,
        pattern: str = None,
        primary_location_id_prefix: str = None,
        secondary_location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Load location crosswalks helper."""
        cache_dir = Path(
            self.ev.dir_path,
            const.CACHE_DIR,
            const.LOADING_CACHE_DIR,
            const.LOCATION_CROSSWALKS_DIR
        )
        # Clear the cache directory if it exists.
        remove_dir_if_exists(cache_dir)

        convert_location_crosswalks(
            in_path,
            cache_dir,
            field_mapping=field_mapping,
            pattern=pattern,
            **kwargs
        )

        # Read the converted files to Spark DataFrame
        df = self._read_files(cache_dir)

        # Add or replace primary location_id prefix if provided
        if primary_location_id_prefix:
            df = add_or_replace_sdf_column_prefix(
                sdf=df,
                column_name="primary_location_id",
                prefix=primary_location_id_prefix,
            )

        # Add or replace secondary location_id prefix if provided
        if secondary_location_id_prefix:
            df = add_or_replace_sdf_column_prefix(
                sdf=df,
                column_name="secondary_location_id",
                prefix=secondary_location_id_prefix,
            )

        # Validate using the validate method
        validated_df = self._validate(df)

        # Write to the table df.rdd.getNumPartitions()
        self._write_spark_df(
            df=validated_df,
            num_partitions=df.rdd.getNumPartitions(),
            write_mode=write_mode,
        )

        # Reload the table
        self._load_table()

    def _get_schema(self, type: str = "pyspark"):
        """Get the location crosswalk schema."""
        if type == "pandas":
            return self.schema_func(type="pandas")

        location_ids = self.ev.locations.distinct_values("id")
        return self.schema_func(location_ids=location_ids)

    def field_enum(self) -> LocationCrosswalkFields:
        """Get the location crosswalk fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return LocationCrosswalkFields(
            "LocationCrosswalkFields",
            {field: field for field in fields}
        )

    def to_pandas(self):
        """Return Pandas DataFrame for Location Crosswalk."""
        self._check_load_table()
        df = self.df.toPandas()
        df.attrs['table_type'] = self.name
        df.attrs['fields'] = self.fields()
        return df

    def to_geopandas(self):
        """Return GeoPandas DataFrame."""
        self._check_load_table()
        gdf = join_geometry(
            self.df, self.ev.locations.to_sdf(),
            "primary_location_id"
        )
        gdf.attrs['table_type'] = self.name
        gdf.attrs['fields'] = self.fields()
        return gdf

    def load_parquet(
        self,
        in_path: Union[Path, str],
        pattern: str = "**/*.parquet",
        field_mapping: dict = None,
        primary_location_id_prefix: str = None,
        secondary_location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import location crosswalks from parquet file format.

        Parameters
        ----------
        in_path : Union[Path, str]
            The input file or directory path.
            Parquet file format.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
        primary_location_id_prefix : str, optional
            The prefix to add to primary location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        secondary_location_id_prefix : str, optional
            The prefix to add to secondary location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are
            overwritten.
        **kwargs
            Additional keyword arguments are passed to pd.read_csv()
            or pd.read_parquet().

        Notes
        -----

        The TEEHR Location Crosswalk table schema includes fields:

        - primary_location_id
        - secondary_location_id
        """
        validate_input_is_parquet(in_path)
        self._load(
            in_path=in_path,
            field_mapping=field_mapping,
            pattern=pattern,
            primary_location_id_prefix=primary_location_id_prefix,
            secondary_location_id_prefix=secondary_location_id_prefix,
            write_mode=write_mode,
            **kwargs
        )
        self._load_table()

    def load_csv(
        self,
        in_path: Union[Path, str],
        pattern: str = "**/*.csv",
        field_mapping: dict = None,
        primary_location_id_prefix: str = None,
        secondary_location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import location crosswalks from CSV file format.

        Parameters
        ----------
        in_path : Union[Path, str]
            The input file or directory path.
            CSV file format.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
        primary_location_id_prefix : str, optional
            The prefix to add to primary location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        secondary_location_id_prefix : str, optional
            The prefix to add to secondary location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are overwritten
        **kwargs
            Additional keyword arguments are passed to pd.read_csv()
            or pd.read_parquet().

        Notes
        -----
        The TEEHR Location Crosswalk table schema includes fields:

        - primary_location_id
        - secondary_location_id
        """
        validate_input_is_csv(in_path)
        self._load(
            in_path=in_path,
            field_mapping=field_mapping,
            pattern=pattern,
            primary_location_id_prefix=primary_location_id_prefix,
            secondary_location_id_prefix=secondary_location_id_prefix,
            write_mode=write_mode,
            **kwargs
        )
        self._load_table()


"""Module for parquet-based query models."""
from collections.abc import Iterable
from typing import List, Union
from pydantic import BaseModel as BaseModel
from pydantic import FieldValidationInfo, field_validator
from datetime import datetime
import logging
from teehr.models.str_enum import StrEnum
from teehr.models.table_enums import (
    ConfigurationFields,
    UnitFields,
    VariableFields,
    AttributeFields,
    LocationFields,
    LocationAttributeFields,
    LocationCrosswalkFields,
    TimeseriesFields,
    JoinedTimeseriesFields
)
from teehr.models.pydantic_table_models import (
    TableBaseModel
)


logger = logging.getLogger(__name__)


class FilterOperators(StrEnum):
    """Filter symbols."""

    eq = "="
    gt = ">"
    lt = "<"
    gte = ">="
    lte = "<="
    islike = "like"
    isin = "in"


class FilterBaseModel(BaseModel):
    """Base model for filters."""

    column: TableBaseModel
    operator: FilterOperators
    value: Union[str, List[str]]

    def is_iterable_not_str(obj):
        """Check if is type Iterable and not str.

        We should not have the case where a string is provided, but doesn't
        hurt to check.
        """
        if isinstance(obj, Iterable) and not isinstance(obj, str):
            return True
        return False

    @field_validator("value")
    def in_operator_must_have_iterable(cls, v, info: FieldValidationInfo):
        """Ensure that an 'in' operator has an iterable type."""
        if cls.is_iterable_not_str(v) and info.data["operator"] != "in":
            raise ValueError("iterable value must be used with 'in' operator")

        if info.data["operator"] == "in" and not cls.is_iterable_not_str(v):
            raise ValueError(
                "'in' operator can only be used with iterable value"
            )
        return v

    @field_validator("column", mode='before')
    def coerce_column_to_enum(cls, v, info: FieldValidationInfo):
        """Column name must exist in the database table."""
        if not isinstance(v, StrEnum):
            fields = info.context.get("fields_enum")
            v = fields[v]
        return v


class ConfigurationFilter(FilterBaseModel):
    """Configuration filter model."""

    column: ConfigurationFields


class UnitFilter(FilterBaseModel):
    """Unit filter model."""

    column: UnitFields


class VariableFilter(FilterBaseModel):
    """Variable filter model."""

    column: VariableFields


class AttributeFilter(FilterBaseModel):
    """Attribute filter model."""

    column: AttributeFields


class LocationFilter(FilterBaseModel):
    """Location filter model.

    ToDo: Add a geometry filter: lat, lon, intersect, etc.
    """

    column: LocationFields


class LocationAttributeFilter(FilterBaseModel):
    """Location attribute filter model.

    ToDo: How should we handle values types that are say a number stored
    as a string?
    """

    column: LocationAttributeFields


class LocationCrosswalkFilter(FilterBaseModel):
    """Location crosswalk filter model."""

    column: LocationCrosswalkFields


class TimeseriesFilter(FilterBaseModel):
    """Timeseries filter model."""

    column: TimeseriesFields
    value: Union[
        str, int, float, datetime, List[Union[str, int, float, datetime]]
    ]


class JoinedTimeseriesFilter(FilterBaseModel):
    """Joined timeseries filter model."""

    column: JoinedTimeseriesFields
    value: Union[
        str, int, float, datetime, List[Union[str, int, float, datetime]]
    ]



from teehr.evaluation.tables.domain_table import DomainTable
from teehr.models.filters import VariableFilter
from teehr.models.table_enums import VariableFields
from teehr.models.pydantic_table_models import Variable
import teehr.models.pandera_dataframe_schemas as schemas
from typing import List, Union
import logging
from teehr.utils.utils import to_path_or_s3path


logger = logging.getLogger(__name__)


class VariableTable(DomainTable):
    """Access methods to variables table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "variables"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.filter_model = VariableFilter
        self.schema_func = schemas.variable_schema
        self.unique_column_set = ["name"]

    def field_enum(self) -> VariableFields:
        """Get the variable fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return VariableFields(
            "VariableFields",
            {field: field for field in fields}
        )

    def add(
        self,
        variable: Union[Variable, List[Variable]]
    ):
        """Add a unit to the evaluation.

        Parameters
        ----------
        variable : Union[Variable, List[Variable]]
            The variable domain to add.

        Example
        -------
        >>> from teehr.models.domain_tables import Variable
        >>> variable = Variable(
        >>>     name="streamflow_hourly_inst",
        >>>     long_name="Instantaneous streamflow"
        >>> )
        >>> ev.load.add_variable(variable)
        """
        self._add(variable)


"""Pydantic models for domain table entries."""
from pydantic import BaseModel
from datetime import datetime


class TableBaseModel(BaseModel):
    """Base model for all tables.

    Adds a class method to get the field names.
    """

    @classmethod
    def get_field_names(cls, by_alias=False) -> list[str]:
        """Get the field names of the model."""
        field_names = []
        for k, v in cls.model_fields.items():
            if by_alias and v.alias:
                field_names.append(v.alias)
            else:
                field_names.append(k)

        return field_names


class Configuration(TableBaseModel):
    """Configuration model."""

    name: str
    type: str
    description: str


class Unit(TableBaseModel):
    """Unit model."""

    name: str
    long_name: str


class Variable(TableBaseModel):
    """Variable model."""

    name: str
    long_name: str


class Attribute(TableBaseModel):
    """Variable model."""

    name: str
    type: str
    description: str


# class Location(TableBaseModel):
#     """Location model."""

#     id: str
#     name: str
#     geometry: bytes


# class LocationAttribute(TableBaseModel):
#     """LocationAttribute model."""

#     location_id: str
#     attribute_name: str
#     value: str


# class LocationCrosswalk(TableBaseModel):
#     """LocationCrosswalk model."""

#     primary_location_id: str
#     secondary_location_id: str


# class Timeseries(TableBaseModel):
#     """Timeseries model."""

#     reference_time: datetime
#     value_time: datetime
#     configuration_name: str
#     unit_name: str
#     variable_name: str
#     value: float
#     location_id: str

# Note: there is no JoinedTimeseries model in this file because the
# JoinedTimeseries model is dynamic and the fields can be different
# for each evaluation.



from teehr.evaluation.tables.base_table import BaseTable
from teehr.models.pydantic_table_models import TableBaseModel
import pandas as pd
from typing import List, Union
import logging

logger = logging.getLogger(__name__)


class DomainTable(BaseTable):
    """Domain table class."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.format = "csv"

    def _add(
        self,
        obj: Union[TableBaseModel, List[TableBaseModel]]
    ):
        logger.info(f"Adding attribute to {self.dir}")
        self._check_load_table()

        org_df = self.to_sdf()

        if issubclass(type(obj), TableBaseModel):
            obj = [obj]

        new_df = self.spark.createDataFrame(pd.DataFrame([o.model_dump() for o in obj]))

        combined_df = org_df.unionByName(new_df).repartition(1)
        validated_df = self._validate(combined_df)

        self._write_spark_df(validated_df)


"""Module for NWM fetching models."""
from teehr.models.str_enum import StrEnum


# Timeseries models.
class TimeseriesTypeEnum(StrEnum):
    """Timeseries types enum."""

    primary = "primary"
    secondary = "secondary"


# Retrospective models.
class ChannelRtRetroVariableEnum(StrEnum):
    """ChannelRtRetro variables enum."""

    nudge = "nudge"
    qBtmVertRunoff = "qBtmVertRunoff"
    qBucket = "qBucket"
    qSfcLatRunoff = "qSfcLatRunoff"
    streamflow = "streamflow"
    velocity = "velocity"


class USGSServiceEnum(StrEnum):
    """USGSServiceEnum for fetching USGS data."""

    dv = "dv"
    iv = "iv"


class USGSChunkByEnum(StrEnum):
    """USGSChunkByEnum for fetching USGS data."""

    day = "day"
    location_id = "location_id"
    week = "week"
    month = "month"
    year = "year"


class NWMChunkByEnum(StrEnum):
    """NWMChunkByEnum for fetching NWM data."""

    week = "week"
    month = "month"
    year = "year"


class SupportedNWMRetroVersionsEnum(StrEnum):
    """SupportedNWMRetroVersionsEnum."""

    nwm20 = "nwm20"
    nwm21 = "nwm21"
    nwm30 = "nwm30"


class SupportedNWMOperationalVersionsEnum(StrEnum):
    """SupportedNWMOperationalVersionsEnum."""

    nwm12 = "nwm12"
    nwm20 = "nwm20"
    nwm21 = "nwm21"
    nwm22 = "nwm22"
    nwm30 = "nwm30"


class SupportedNWMDataSourcesEnum(StrEnum):
    """SupportedNWMDataSourcesEnum."""

    GCS = "GCS"
    NOMADS = "NOMADS"
    DSTOR = "DSTOR"


class SupportedKerchunkMethod(StrEnum):
    """SupportedKerchunkMethod."""

    local = "local"
    remote = "remote"
    auto = "auto"


class SupportedNWMRetroDomainsEnum(StrEnum):
    """SupportedNWMRetroDomainsEnum."""

    CONUS = "CONUS"
    Alaska = "Alaska"
    PR = "PR"
    Hawaii = "Hawaii"

# class NWM20RetroConfig(BaseModel):
#     URL = 's3://noaa-nwm-retro-v2-zarr-pds'
#     MIN_DATE = datetime(1993, 1, 1)
#     MAX_DATE = datetime(2018, 12, 31, 23)


# class NWM21RetroConfig(BaseModel):
#     URL = "s3://noaa-nwm-retrospective-2-1-zarr-pds/chrtout.zarr/"
#     MIN_DATE = pd.Timestamp(1979, 1, 1)
#     MAX_DATE = pd.Timestamp(2020, 12, 31, 23)



"""Field enums for each table in the dataset."""
from teehr.models.str_enum import StrEnum


class TableWriteEnum(StrEnum):
    """Methods to write or update Evaluation tables."""

    append = "append"
    upsert = "upsert"
    overwrite = "overwrite"


class ConfigurationFields(StrEnum):
    """Empty class for ConfigurationFieldEnum."""

    pass


class UnitFields(StrEnum):
    """Empty class for UnitFieldEnum."""

    pass


class VariableFields(StrEnum):
    """Empty class for VariableFieldEnum."""

    pass


class AttributeFields(StrEnum):
    """Empty class for AttributeFieldEnum."""

    pass


class LocationFields(StrEnum):
    """Empty class for LocationFieldEnum."""

    pass


class LocationAttributeFields(StrEnum):
    """Empty class for LocationAttributeFieldEnum."""

    pass


class LocationCrosswalkFields(StrEnum):
    """Empty class for LocationCrosswalkFieldEnum."""

    pass


class TimeseriesFields(StrEnum):
    """Empty class for TimeseriesFieldEnum."""

    pass


class JoinedTimeseriesFields(StrEnum):
    """Empty class for JoinedTimeseriesFieldEnum."""

    pass



"""Classes representing UDFs."""
from typing import List, Union
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, ConfigDict
import pandas as pd
import pyspark.sql.types as T
from pyspark.sql.functions import pandas_udf
import pyspark.sql as ps
from teehr.models.calculated_fields.base import CalculatedFieldABC
from teehr.models.calculated_fields.base import CalculatedFieldBaseModel


class Month(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds the month from a timestamp column.

    Properties
    ----------

    - input_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - output_field_name:
        The name of the column to store the month.
        Default: "month"
    """

    input_field_name: str = Field(
        default="value_time"
    )
    output_field_name: str = Field(
        default="month"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.IntegerType())
        def func(col: pd.Series) -> pd.Series:
            return col.dt.month

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.input_field_name)
        )
        return sdf


class Year(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds the year from a timestamp column.

    Properties
    ----------
    - input_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - output_field_name:
        The name of the column to store the year.
        Default: "year"

    """

    input_field_name: str = Field(
        default="value_time"
    )
    output_field_name: str = Field(
        default="year"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.IntegerType())
        def func(col: pd.Series) -> pd.Series:
            return col.dt.year

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.input_field_name)
        )
        return sdf


class WaterYear(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds the water year from a timestamp column.

    Properties
    ----------
    - input_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - output_field_name:
        The name of the column to store the water year.
        Default: "water_year"

    Water year is defined as the year of the date plus one if the month is
    October or later.
    """

    input_field_name: str = Field(
        default="value_time"
    )
    output_field_name: str = Field(
        default="water_year"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.IntegerType())
        def func(col: pd.Series) -> pd.Series:
            return col.dt.year + (col.dt.month >= 10).astype(int)

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.input_field_name)
        )
        return sdf


class NormalizedFlow(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Normalize flow values by drainage area.

    Properties
    ----------
    - primary_value_field_name:
        The name of the column containing the flow values.
        Default: "primary_value"
    - drainage_area_field_name:
        The name of the column containing the drainage area.
        Default: "drainage_area"
    - output_field_name:
        The name of the column to store the normalized flow values.
        Default: "normalized_flow"

    """

    primary_value_field_name: str = Field(
        default="primary_value"
    )
    drainage_area_field_name: str = Field(
        default="drainage_area"
    )
    output_field_name: str = Field(
        default="normalized_flow"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.FloatType())
        def func(value: pd.Series, area: pd.Series) -> pd.Series:
            return value.astype(float) / area.astype(float)

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.primary_value_field_name, self.drainage_area_field_name)
        )
        return sdf


class Seasons(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds the season from a timestamp column.

    Properties
    ----------
    - value_time_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - season_months:
        A dictionary mapping season names to the months that define them.

        .. code-block:: python

            Default: {
                "winter": [12, 1, 2],
                "spring": [3, 4, 5],
                "summer": [6, 7, 8],
                "fall": [9, 10, 11]
            }

    - output_field_name:
        The name of the column to store the season.
        Default: "season"

    """

    value_time_field_name: str = Field(
        default="value_time"
    )
    season_months: dict = Field(
        default={
            "winter": [12, 1, 2],
            "spring": [3, 4, 5],
            "summer": [6, 7, 8],
            "fall": [9, 10, 11]
        }
    )
    output_field_name: str = Field(
        default="season"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.StringType())
        def func(value_time: pd.Series) -> pd.Series:
            return value_time.dt.month.apply(
                lambda x: next(
                    (season for season, months in self.season_months.items() if x in months),
                    None
                )
            )

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.value_time_field_name)
        )
        return sdf


class ForecastLeadTime(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds the forecast lead time from a timestamp column.

    Properties
    ----------
    - value_time_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - reference_time_field_name:
        The name of the column containing the forecast time.
        Default: "reference_time"
    - output_field_name:
        The name of the column to store the forecast lead time.
        Default: "forecast_lead_time"

    """

    value_time_field_name: str = Field(
        default="value_time"
    )
    reference_time_field_name: str = Field(
        default="reference_time"
    )
    output_field_name: str = Field(
        default="forecast_lead_time"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.DayTimeIntervalType())
        def func(value_time: pd.Series,
                 reference_time: pd.Series
                 ) -> pd.Series:
            difference = value_time - reference_time
            return difference

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.value_time_field_name, self.reference_time_field_name)
        )
        return sdf


class ThresholdValueExceeded(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds boolean column indicating if the input value exceeds a threshold.

    Properties
    ----------
    - input_field_name:
        The name of the column containing the primary value.
        Default: "primary_value"
    - threshold_field_name:
        The name of the column containing the threshold value.
        Default: 0
    - output_field_name:
        The name of the column to store the boolean value.
        Default: "threshold_value_exceeded"

    """

    input_field_name: str = Field(
        default="primary_value"
    )
    threshold_field_name: str = Field(
        default="secondary_value"
    )
    output_field_name: str = Field(
        default="threshold_value_exceeded"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.BooleanType())
        def func(input_value: pd.Series,
                 threshold_value: pd.Series
                 ) -> pd.Series:
            mask = input_value > threshold_value
            return mask

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.input_field_name,
                 self.threshold_field_name)
        )
        return sdf


class DayOfYear(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds the day of the year from a timestamp column.

    Properties
    ----------
    - input_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - output_field_name:
        The name of the column to store the day of the year.
        Default: "day_of_year"

    Notes
    -----
    - February 29th in leap years is set to None.
    - All days after February 29th are adjusted to correspond to the same day
      of the year as in a non-leap year.
    """

    input_field_name: str = Field(
        default="value_time"
    )
    output_field_name: str = Field(
        default="day_of_year"
    )

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:
        """Apply the calculated field to the Spark DataFrame."""
        @pandas_udf(returnType=T.IntegerType())
        def func(col: pd.Series) -> pd.Series:
            def adjust_day_of_year(date):
                if (date.year % 4 == 0 and date.year % 100 != 0) or (date.year % 400 == 0):
                    if date.month == 2 and date.day == 29:
                        return None
                    elif date.month > 2 or (date.month == 2 and date.day > 28):
                        return date.dayofyear - 1
                    else:
                        return date.dayofyear
                else:
                    return date.dayofyear

            return col.apply(adjust_day_of_year)

        sdf = sdf.withColumn(
            self.output_field_name,
            func(self.input_field_name)
        )
        return sdf


class RowLevelCalculatedFields:
    """Row level Calculated Fields.

    Notes
    -----
    Row level CFs are applied to each row in the table based on data that is
    in one or more existing fields.  These are applied per row and are not
    aware of the data in any other row (e.g., are not aware of any other
    timeseries values in a "timeseries").  This can be used for adding fields
    such as a field based on the data/time (e.g., month, year, season, etc.)
    or based on the value field (e.g., normalized flow, log flow, etc.) and
    many other uses.

    Available Calculated Fields:

    - Month
    - Year
    - WaterYear
    - NormalizedFlow
    - Seasons
    - ForecastLeadTime
    - ThresholdValueExceeded
    - DayOfYear
    """

    Month = Month
    Year = Year
    WaterYear = WaterYear
    NormalizedFlow = NormalizedFlow
    Seasons = Seasons
    ForecastLeadTime = ForecastLeadTime
    ThresholdValueExceeded = ThresholdValueExceeded
    DayOfYear = DayOfYear



"""Classes representing available performance metrics."""
from typing import List, Dict, Callable, Union

from pydantic import Field
import teehr.models.metrics.metric_attributes as tma
from teehr.models.metrics.basemodels import (
    TransformEnum,
    DeterministicBasemodel,
    BootstrapBasemodel
)
from teehr.metrics import deterministic_funcs as metric_funcs
from teehr.models.str_enum import StrEnum


class MeanError(DeterministicBasemodel):
    """Mean Error.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "mean_error".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.mean_error.`
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="mean_error")
    func: Callable = Field(metric_funcs.me_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.ME_ATTRS, frozen=True)


class RelativeBias(DeterministicBasemodel):
    """Relative Bias.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "relative_bias".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.relative_bias`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="relative_bias")
    func: Callable = Field(metric_funcs.rb_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.RBIAS_ATTRS, frozen=True)


class MultiplicativeBias(DeterministicBasemodel):
    """Multiplicative Bias.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "multiplicative_bias".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.multiplicative_bias.`
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="multiplicative_bias")
    func: Callable = Field(metric_funcs.mb_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.MULTBIAS_ATTRS, frozen=True)


class MeanSquareError(DeterministicBasemodel):
    """Mean Square Error.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "mean_squared_error".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.mean_squared_error`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="mean_square_error")
    func: Callable = Field(metric_funcs.mse_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.MSE_ATTRS, frozen=True)


class RootMeanSquareError(DeterministicBasemodel):
    """Root Mean Squared Error.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "root_mean_square_error".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.root_mean_square_error`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="root_mean_square_error")
    func: Callable = Field(metric_funcs.rmse_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.RMSE_ATTRS, frozen=True)


class MeanAbsoluteError(DeterministicBasemodel):
    """Mean Absolute Error.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "mean_absolute_error".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.mean_absolute_error`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="mean_absolute_error")
    func: Callable = Field(metric_funcs.mae_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.MAE_ATTRS, frozen=True)


class MeanAbsoluteRelativeError(DeterministicBasemodel):
    """Relative Mean Absolute Error.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "mean_absolute_relative_error".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.mean_absolute_relative_error`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="mean_absolute_relative_error")
    func: Callable = Field(metric_funcs.mare_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.RMAE_ATTRS, frozen=True)


class PearsonCorrelation(DeterministicBasemodel):
    """Pearson Correlation.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "pearson_correlation".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.pearson_correlation`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="pearson_correlation")
    func: Callable = Field(metric_funcs.pc_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.PEARSON_ATTRS, frozen=True)


class Rsquared(DeterministicBasemodel):
    """Coefficient of Determination.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "r_squared".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.r_squared`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="r_squared")
    func: Callable = Field(metric_funcs.r_squared_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.R2_ATTRS, frozen=True)


class NashSutcliffeEfficiency(DeterministicBasemodel):
    """Nash-Sutcliffe Efficiency.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "nash_sutcliffe_efficiency".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.nash_sutcliffe_efficiency`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="nash_sutcliffe_efficiency")
    func: Callable = Field(metric_funcs.nse_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.NSE_ATTRS, frozen=True)


class NormalizedNashSutcliffeEfficiency(DeterministicBasemodel):
    """Normalized Nash-Sutcliffe Efficiency.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default
        "nash_sutcliffe_efficiency_normalized".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.nash_sutcliffe_efficiency_normalized`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(
        default="nash_sutcliffe_efficiency_normalized"
    )
    func: Callable = Field(metric_funcs.nse_norm_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.NNSE_ATTRS, frozen=True)


class KlingGuptaEfficiency(DeterministicBasemodel):
    """Kling-Gupta Efficiency.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "kling_gupta_efficiency".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.kling_gupta_efficiency`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    sr : float
        The scaling factor for the correlation component, by default 1.0.
    sa : float
        The scaling factor for the variability component, by default 1.0.
    sb : float
        The scaling factor for the bias component, by default 1.0.
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="kling_gupta_efficiency")
    func: Callable = Field(metric_funcs.kge_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    sr: float = Field(default=1.0)
    sa: float = Field(default=1.0)
    sb: float = Field(default=1.0)
    attrs: Dict = Field(default=tma.KGE_ATTRS, frozen=True)


class KlingGuptaEfficiencyMod1(DeterministicBasemodel):
    """Kling-Gupta Efficiency - modified 1 (2012).

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "kling_gupta_efficiency_mod1".
    func : Callable
        The function to apply to the data, by default
        :func:`kling_gupta_efficiency_mod1`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    sr : float
        The scaling factor for the correlation component, by default 1.0.
    sa : float
        The scaling factor for the variability component, by default 1.0.
    sb : float
        The scaling factor for the bias component, by default 1.0.
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="kling_gupta_efficiency_mod1")
    func: Callable = Field(metric_funcs.kge_mod1_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    sr: float = Field(default=1.0)
    sa: float = Field(default=1.0)
    sb: float = Field(default=1.0)
    attrs: Dict = Field(default=tma.KGE1_ATTRS, frozen=True)


class KlingGuptaEfficiencyMod2(DeterministicBasemodel):
    """Kling-Gupta Efficiency - modified 2 (2021).

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "kling_gupta_efficiency_mod2".
    func : Callable
        The function to apply to the data, by default
        :func:`kling_gupta_efficiency_mod2`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    sr : float
        The scaling factor for the correlation component, by default 1.0.
    sa : float
        The scaling factor for the variability component, by default 1.0.
    sb : float
        The scaling factor for the bias component, by default 1.0.
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="kling_gupta_efficiency_mod2")
    func: Callable = Field(metric_funcs.kge_mod2_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    sr: float = Field(default=1.0)
    sa: float = Field(default=1.0)
    sb: float = Field(default=1.0)
    attrs: Dict = Field(default=tma.KGE2_ATTRS, frozen=True)


class SpearmanCorrelation(DeterministicBasemodel):
    """Spearman Rank Correlation Coefficient.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "spearman_correlation".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.spearman_correlation`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="spearman_correlation")
    func: Callable = Field(metric_funcs.spearman_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.SPEARMAN_R_ATTRS, frozen=True)


class MaxValueDelta(DeterministicBasemodel):
    """Max Value Delta.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "max_value_delta".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.max_value_delta`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="max_value_delta")
    func: Callable = Field(metric_funcs.mvd_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.MAX_VALUE_DELTA_ATTRS, frozen=True)


class MaxValueTimeDelta(DeterministicBasemodel):
    """Max Value Time Delta.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "max_value_timedelta".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.max_value_timedelta`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default
        ["primary_value", "secondary_value", "value_time"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="max_value_time_delta")
    func: Callable = Field(metric_funcs.mvtd_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value", "value_time"]
    )
    attrs: Dict = Field(default=tma.MAX_VALUE_TIMEDELTA_ATTRS, frozen=True)


class AnnualPeakRelativeBias(DeterministicBasemodel):
    """Annual Peak Relative Bias.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "annual_peak_relative_bias".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.annual_peak_relative_bias`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default
        ["primary_value", "secondary_value", "value_time"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="annual_peak_flow_bias")
    func: Callable = Field(metric_funcs.aprb_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value", "value_time"]
    )
    attrs: Dict = Field(default=tma.ANNUAL_PEAK_RBIAS_ATTRS, frozen=True)


class RootMeanStandardDeviationRatio(DeterministicBasemodel):
    """Root Mean Standard Deviation Ratio.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "root_mean_standard_deviation_ratio".
    func : Callable
        The function to apply to the data, by default
        :func:`deterministic_funcs.root_mean_standard_deviation_ratio`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default
        ["primary_value", "secondary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="root_mean_standard_deviation_ratio")  # noqa: E501
    func: Callable = Field(metric_funcs.rmsdr_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value"]
    )
    attrs: Dict = Field(default=tma.RSR_ATTRS, frozen=True)


class DeterministicMetrics:
    """Define and customize determinisitic metrics.

    Notes
    -----
    Deterministic metrics compare two timeseries, typically primary ("observed")
    vs. secondary ("modeled") values. Available metrics include:

    - AnnualPeakRelativeBias
    - KlingGuptaEfficiency
    - KlingGuptaEfficiencyMod1
    - KlingGuptaEfficiencyMod2
    - MaxValueDelta
    - MaxValueTimeDelta
    - MeanError
    - MeanAbsoluteError
    - MeanAbsoluteRelativeError
    - MeanSquareError
    - MultiplicativeBias
    - PearsonCorrelation
    - NashSutcliffeEfficiency
    - NormalizedNashSutcliffeEfficiency
    - RelativeBias
    - RootMeanSquareError
    - Rsquared
    - SpearmanCorrelation
    - RootMeanStandardDeviationRatio
    """

    AnnualPeakRelativeBias = AnnualPeakRelativeBias
    KlingGuptaEfficiency = KlingGuptaEfficiency
    KlingGuptaEfficiencyMod1 = KlingGuptaEfficiencyMod1
    KlingGuptaEfficiencyMod2 = KlingGuptaEfficiencyMod2
    MaxValueDelta = MaxValueDelta
    MaxValueTimeDelta = MaxValueTimeDelta
    MeanError = MeanError
    MeanAbsoluteError = MeanAbsoluteError
    MeanAbsoluteRelativeError = MeanAbsoluteRelativeError
    MeanSquareError = MeanSquareError
    MultiplicativeBias = MultiplicativeBias
    PearsonCorrelation = PearsonCorrelation
    NashSutcliffeEfficiency = NashSutcliffeEfficiency
    NormalizedNashSutcliffeEfficiency = NormalizedNashSutcliffeEfficiency
    RelativeBias = RelativeBias
    RootMeanSquareError = RootMeanSquareError
    Rsquared = Rsquared
    SpearmanCorrelation = SpearmanCorrelation
    RootMeanStandardDeviationRatio = RootMeanStandardDeviationRatio



import sys
from pathlib import Path
from teehr.evaluation.tables.timeseries_table import TimeseriesTable
from teehr.models.filters import JoinedTimeseriesFilter
from teehr.models.table_enums import JoinedTimeseriesFields, TableWriteEnum
from teehr.querying.utils import join_geometry
import teehr.models.pandera_dataframe_schemas as schemas
import pyspark.sql as ps
import logging
from teehr.utils.utils import to_path_or_s3path
from teehr.models.calculated_fields.base import CalculatedFieldBaseModel
from typing import List, Union

logger = logging.getLogger(__name__)


class JoinedTimeseriesTable(TimeseriesTable):
    """Access methods to joined timeseries table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "joined_timeseries"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.filter_model = JoinedTimeseriesFilter
        self.validate_filter_field_types = False
        self.strict_validation = False
        self.schema_func = schemas.joined_timeseries_schema
        self.unique_column_set = [
            "primary_location_id",
            "secondary_location_id",
            "value_time",
            "reference_time",
            "variable_name",
            "unit_name",
        ]

    def field_enum(self) -> JoinedTimeseriesFields:
        """Get the joined timeseries fields enum."""
        self._check_load_table()
        fields_list = self.df.columns
        return JoinedTimeseriesFields(
            "JoinedTimeseriesFields",
            {field: field for field in fields_list}
        )

    def to_pandas(self):
        """Return Pandas DataFrame for Joined Timeseries."""
        self._check_load_table()
        df = self.df.toPandas()
        df.attrs['table_type'] = 'joined_timeseries'
        df.attrs['fields'] = self.fields()
        return df

    def to_geopandas(self):
        """Return GeoPandas DataFrame."""
        self._check_load_table()
        return join_geometry(
            self.df, self.ev.locations.to_sdf(),
            "primary_location_id"
        )

    def _join(self) -> ps.DataFrame:
        """Join primary and secondary timeseries."""
        joined_df = self.ev.sql("""
            SELECT
                sf.reference_time
                , sf.value_time as value_time
                , pf.location_id as primary_location_id
                , sf.location_id as secondary_location_id
                , pf.value as primary_value
                , sf.value as secondary_value
                , sf.configuration_name
                , sf.unit_name
                , sf.variable_name
                , sf.member
            FROM secondary_timeseries sf
            JOIN location_crosswalks cf
                on cf.secondary_location_id = sf.location_id
            JOIN primary_timeseries pf
                on cf.primary_location_id = pf.location_id
                and sf.value_time = pf.value_time
                and sf.unit_name = pf.unit_name
                and sf.variable_name = pf.variable_name
        """,
        create_temp_views=["secondary_timeseries", "location_crosswalks", "primary_timeseries"])
        return joined_df

    def _add_attr(self, joined_df: ps.DataFrame) -> ps.DataFrame:
        """Add attributes to the joined timeseries dataframe."""

        location_attributes_df = self.ev.location_attributes.to_sdf()

        joined_df.createTempView("joined")

        # Get distinct attribute names
        distinct_attributes = self.ev.location_attributes.distinct_values("attribute_name")

        # Pivot the table
        pivot_df = (
            location_attributes_df.groupBy("location_id")
            .pivot("attribute_name", distinct_attributes).agg({"value": "max"})
        )

        # Create a view
        pivot_df.createTempView("attrs")

        joined_df = self.spark.sql("""
            SELECT
                joined.*
                , attrs.*
            FROM joined
            JOIN attrs
                on joined.primary_location_id = attrs.location_id
        """)

        self.spark.catalog.dropTempView("joined")
        self.spark.catalog.dropTempView("attrs")

        return joined_df

    def add_calculated_fields(self, cfs: Union[CalculatedFieldBaseModel, List[CalculatedFieldBaseModel]]):
        """Add calculated fields to the joined timeseries table.

        Note this does not persist the CFs to the table. It only applies them to the DataFrame.
        To persist the CFs to the table, use the `write` method.

        Parameters
        ----------
        cfs : Union[CalculatedFieldBaseModel, List[CalculatedFieldBaseModel]]
            The CFs to apply to the DataFrame.

        Examples
        --------
        >>> import teehr
        >>> from teehr import RowLevelCalculatedFields as rcf
        >>> ev.join_timeseries.add_calculated_fields([
        >>>     rcf.Month()
        >>> ]).write()
        """
        self._check_load_table()

        if not isinstance(cfs, List):
            cfs = [cfs]

        for cf in cfs:
            self.df = cf.apply_to(self.df)

        return self

    def write(self):
        """Write the joined timeseries table to disk."""
        self._write_spark_df(self.df)
        logger.info("Joined timeseries table written to disk.")
        self._load_table()

    def _run_script(self, joined_df: ps.DataFrame) -> ps.DataFrame:
        """Add UDFs to the joined timeseries dataframe."""
        try:
            sys.path.append(str(Path(self.ev.scripts_dir).resolve()))
            import user_defined_fields as udf # noqa
            joined_df = udf.add_user_defined_fields(joined_df)
        except ImportError:
            logger.info(
                f"No user-defined fields found in {self.ev.scripts_dir}."
                "Not adding user-defined fields."
            )
            return joined_df

        return joined_df

    def create(
        self,
        add_attrs: bool = False,
        execute_scripts: bool = False,
    ):
        """Create joined timeseries table.

        Parameters
        ----------
        execute_scripts : bool, optional
            Execute UDFs, by default False
        add_attrs : bool, optional
            Add attributes, by default False
        """
        joined_df = self._join()

        if add_attrs:
            joined_df = self._add_attr(joined_df)

        if execute_scripts:
            joined_df = self._run_script(joined_df)

        validated_df = self._validate(joined_df, False)
        self._write_spark_df(validated_df)
        logger.info("Joined timeseries table created.")
        self._load_table()

    def _check_for_null_partition_by_values(self, df: ps.DataFrame) -> ps.DataFrame:
        """Remove a field from partition_by if all values are null."""
        partition_by = self.partition_by
        if partition_by is None:
            partition_by = []
        for field_name in partition_by:
            if field_name in df.columns:
                if len(df.filter(df[field_name].isNotNull()).collect()) == 0:
                    logger.debug(
                        f"All {field_name} values are null. "
                        f"{field_name} will be set to a default value."
                    )
                    self.partition_by.remove(field_name)
        return df

    def _write_spark_df(
        self,
        df: ps.DataFrame,
        **kwargs
    ):
        """Write spark dataframe to directory in overwrite mode.

        Parameters
        ----------
        df : ps.DataFrame
            The spark dataframe to write.
        **kwargs
            Additional options to pass to the spark write method.
        """
        if self.ev.is_s3:
            logger.error("Writing to S3 is not supported.")
            raise ValueError("Writing to S3 is not supported.")

        logger.info(f"Writing files to {self.dir}.")

        self._check_for_null_partition_by_values(df)

        if len(kwargs) == 0:
            kwargs = {
                "header": "true",
            }
        partition_by = self.partition_by
        if partition_by is None:
            partition_by = []
        (
            df.
            write.
            partitionBy(partition_by).
            format(self.format).
            mode("overwrite").
            options(**kwargs).
            save(str(self.dir))
        )



"""Module for importing locations from a file."""
from typing import Union
from pathlib import Path
import teehr.loading.utils as tpu
import teehr.models.pandera_dataframe_schemas as schemas
from teehr.loading.utils import merge_field_mappings
import teehr.models.pandera_dataframe_schemas as schemas
import logging

logger = logging.getLogger(__name__)


def convert_single_locations(
        input_filepath: Union[str, Path],
        output_filepath: Union[str, Path],
        field_mapping: dict,
        **kwargs
):
    """Import locations from a file.

    This function reads a geodata file, formats for TEEHR,
    converts to geoparquet and writes to disk..

    Parameters
    ----------
    input_filepath : Union[str, Path]
        The input file path.
    output_filepath : Union[str, Path]
        The output file path.
    field_mapping : dict, optional
        A dictionary mapping input fields to output fields.
        format: {input_field: output_field}
    **kwargs
        Additional keyword arguments are passed to GeoPandas read_file().

    Returns
    -------
    None

    field_mapping = {"id_field": "id", "name_field": "name"}
    Note, not consistent with the field_mapping in pre/timeseries.py

    Future: Add support for writing attr to attr table.
    """
    logger.info(
        f"Converting locations from {input_filepath}."
    )

    # read file
    locations = tpu.read_spatial_file(input_filepath, **kwargs)

    # convert to EPSG:4326
    locations.to_crs("EPSG:4326", inplace=True)

    locations.rename(columns=field_mapping, inplace=True)

    schema = schemas.locations_schema(type="pandas")
    validated_df = schema.validate(locations)

    # write to parquet
    output_filepath = Path(output_filepath)
    output_filepath.parent.mkdir(parents=True, exist_ok=True)
    validated_df.to_parquet(output_filepath)

    logger.info(f"Locations written to {output_filepath}.")


def convert_locations(
    in_path: Union[str, Path],
    out_dirpath: Union[str, Path],
    field_mapping: dict = None,
    pattern: str = "**/*.parquet",
    **kwargs
):
    """Convert timeseries data to parquet format.

    Parameters
    ----------
    in_path : Union[str, Path]
        The input file path.
    out_dirpath : Union[str, Path]
        The output file path.
    field_mapping : dict, optional
        A dictionary mapping input fields to output fields.
        format: {input_field: output_field}
    pattern : str, optional (default: "**/*.parquet")
        The pattern to match files.
    **kwargs
        Additional keyword arguments are passed to GeoPandas read_file().

    Can convert any file format that GeoPandas can read.
    """
    in_path = Path(in_path)
    out_dirpath = Path(out_dirpath)

    default_field_mapping = {}
    for field in schemas.locations_schema(type="pandas").columns.keys():
        if field not in default_field_mapping.values():
            default_field_mapping[field] = field

    if field_mapping:
        logger.debug("Merging user field_mapping with default field mapping.")
        field_mapping = merge_field_mappings(
            default_field_mapping,
            field_mapping
        )
    else:
        logger.debug("Using default field mapping.")
        field_mapping = default_field_mapping

    files_converted = 0
    if in_path.is_dir():
        # recursively convert all files in directory
        logger.info(f"Recursively converting all files in {in_path}/{pattern}")
        for in_filepath in in_path.glob(f"{pattern}"):
            relative_name = in_filepath.relative_to(in_path)
            out_filepath = Path(out_dirpath, relative_name)
            out_filepath = out_filepath.with_suffix(".parquet")
            convert_single_locations(
                in_filepath,
                out_filepath,
                field_mapping,
                **kwargs
            )
            files_converted += 1
    else:
        logger.info(f"Converting locations data: {in_path}")
        out_filepath = Path(out_dirpath, in_path.name)
        out_filepath = out_filepath.with_suffix(".parquet")
        convert_single_locations(
            in_path,
            out_filepath,
            field_mapping,
            **kwargs
        )
        files_converted += 1
    logger.info(f"Converted {files_converted} files.")



"""Test script for debugging evaluation data."""
from teehr import Evaluation
from pathlib import Path
from teehr import Metrics as metrics
from teehr.models.table_enums import (
    JoinedTimeseriesFields
)

# Set a path to the directory where the evaluation will be created
TEST_STUDY_DIR = Path(Path().home(), "temp", "real_study")
TEST_STUDY_DIR.mkdir(parents=True, exist_ok=True)

# Create an Evaluation object
ev = Evaluation(dir_path=TEST_STUDY_DIR)

# Enable logging
ev.enable_logging()

fields_list = ev.joined_timeseries.fields()
fields_list.append("relative_bias")

jtf = JoinedTimeseriesFields(
    "JoinedTimeseriesFields",
    {field: field for field in fields_list}
)

# This does not work.
df = (
    ev.metrics.query(
        order_by=["primary_location_id", "month"],
        group_by=["primary_location_id", "month"],
        include_metrics=[
            metrics.KlingGuptaEfficiency(),
            metrics.NashSutcliffeEfficiency(),
            metrics.RelativeBias()
        ]
    )
    .query(
        order_by=["primary_location_id"],
        group_by=["primary_location_id"],
        include_metrics=[
            metrics.PrimaryAverage(
                input_field_names=[jtf.relative_bias],
            )
        ]
    )
    .to_pandas()
)
print(df)


"""Base class for tables."""

from teehr.models.str_enum import StrEnum
from teehr.querying.filter_format import validate_and_apply_filters
import pyspark.sql as ps
from typing import List, Union
from pathlib import Path
from teehr.querying.utils import order_df
from teehr.utils.s3path import S3Path
from teehr.utils.utils import to_path_or_s3path, path_to_spark
from teehr.models.filters import FilterBaseModel
from teehr.models.table_enums import TableWriteEnum
import logging
from pyspark.sql.functions import lit, col, row_number, asc
from pyspark.sql.window import Window
import shutil


logger = logging.getLogger(__name__)


class BaseTable():
    """Base table class."""

    def __init__(self, ev):
        """Initialize class."""
        self.ev = ev
        self.name = None
        self.dir = None
        self.schema_func = None
        self.format = None
        self.partition_by = None
        self.spark = ev.spark
        self.df: ps.DataFrame = None
        self.filter_model: FilterBaseModel = None
        self.strict_validation = True
        self.validate_filter_field_types = True

    @staticmethod
    def _raise_missing_table_error(table_name: str):
        """Raise an error if the table does not exist."""
        err_msg = (
            f"The '{table_name}' table does not exist in the dataset."
            " Please load it first."
        )
        logger.error(err_msg)
        raise ValueError(err_msg)

    def _drop_duplicates(
        self,
        sdf: ps.DataFrame,
    ) -> ps.DataFrame:
        """Drop duplicates from the DataFrame."""
        logger.info(f"Dropping duplicates from {self.name}.")
        window_spec = Window.partitionBy(*self.unique_column_set)
        ordering_expr = [asc(col) for col in self.unique_column_set]
        window_spec = window_spec.orderBy(*ordering_expr)
        sdf_with_row_num = sdf.withColumn("row_num", row_number().over(window_spec))
        deduplicated_sdf = sdf_with_row_num.filter("row_num == 1").drop("row_num")
        return deduplicated_sdf

    def _read_files(
        self,
        path: Union[str, Path, S3Path],
        pattern: str = None,
        use_table_schema: bool = False,
        show_missing_table_warning: bool = True,
        **options
    ) -> ps.DataFrame:
        """Read data from table directory as a spark dataframe.

        Parameters
        ----------
        path : Union[str, Path, S3Path]
            The path to the directory containing the files.
        pattern : str, optional
            The pattern to match files.
        use_table_schema : bool, optional
            If True, use the table schema to read the files.
            Missing files will be ignored with 'ignoreMissingFiles'
            set to True (default).
        show_missing_table_warning : bool, optional
            If True, show the warning an empty table was returned.
            The default is True.
        **options
            Additional options to pass to the spark read method.

        Returns
        -------
        df : ps.DataFrame
            The spark dataframe.
        """
        logger.info(f"Reading files from {path}.")
        if len(options) == 0:
            options = {
                "header": "true",
                "ignoreMissingFiles": "true"
            }

        path = to_path_or_s3path(path)

        path = path_to_spark(path, pattern)

        if use_table_schema is True:
            schema = self.schema_func().to_structtype()
            df = self.ev.spark.read.format(self.format).options(**options).load(path, schema=schema)
            if len(df.head(1)) == 0 and show_missing_table_warning:
                logger.warning(f"An empty dataframe was returned for '{self.name}'.")
        else:
            df = self.ev.spark.read.format(self.format).options(**options).load(path)

        return df

    def _load_table(self, **kwargs):
        """Load the table from the directory to self.df.

        Parameters
        ----------
        **kwargs
            Additional options to pass to the spark read method.
        """
        logger.info(f"Loading files from {self.dir}.")
        self.df = self._read_files(self.dir, **kwargs)

    def _check_load_table(self):
        """Check if the table is loaded.

        If the table is not loaded, try to load it.  If the table is still
        not loaded, raise an error.
        """
        if self.df is None:
            self._load_table()
        if self.df is None:
            self._raise_missing_table_error(table_name=self.name)

    def _upsert_without_duplicates(
        self,
        df,
        num_partitions: int = None,
        **kwargs
    ):
        """Update existing data and append new data without duplicates."""
        logger.info(
            f"Upserting to {self.name} without duplicates."
        )
        partition_by = self.partition_by
        if partition_by is None:
            partition_by = []

        existing_sdf = self._read_files(
            self.dir,
            use_table_schema=True,
            show_missing_table_warning=False,
            **kwargs
        )

        # Limit the dataframe to the partitions that are being updated.
        for partition in partition_by:
            partition_values = df.select(partition).distinct(). \
                rdd.flatMap(lambda x: x).collect()
            if partition_values[0] is not None:  # all null partition values
                existing_sdf = existing_sdf.filter(
                    col(partition).isin(partition_values)
                )

        # Remove rows from existing_sdf that are to be updated.
        # Concat and re-write.
        if not existing_sdf.isEmpty():
            existing_sdf = existing_sdf.join(
                df,
                how="left_anti",
                on=self.unique_column_set,
            )
            df = existing_sdf.unionByName(df)
            # Get columns in correct order
            df = df.select([*self.schema_func().columns])

        # Re-validate since the table was changed
        validated_df = self._validate(df)

        # Drop potential duplicates
        validated_df = self._drop_duplicates(validated_df)

        if num_partitions is not None:
            validated_df = validated_df.repartition(num_partitions)
        (
            validated_df.
            write.
            partitionBy(partition_by).
            format(self.format).
            mode("overwrite").
            options(**kwargs).
            save(str(self.dir))
        )

    def _append_without_duplicates(
        self,
        df,
        num_partitions: int = None,
        **kwargs
    ):
        """Append new data without duplicates."""
        logger.info(
            f"Appending to {self.name} without duplicates."
        )
        partition_by = self.partition_by
        if partition_by is None:
            partition_by = []

        existing_sdf = self._read_files(
            self.dir,
            use_table_schema=True,
            show_missing_table_warning=False,
            **kwargs
        )

        # existing_sdf = existing_sdf.persist()

        # Anti-join: Joins rows from left df that do not have a match
        # in right df.  This is used to drop duplicates. df gets written
        # in append mode.
        if not existing_sdf.isEmpty():
            df = df.join(
                existing_sdf,
                how="left_anti",
                on=self.unique_column_set,
            )

        # Only continue if there is new data to write.
        if not df.isEmpty():
            # Re-validate since the table was changed
            validated_df = self._validate(df)

            if num_partitions is not None:
                validated_df = validated_df.repartition(num_partitions)

            # Drop potential duplicates
            validated_df = self._drop_duplicates(validated_df)

            (
                validated_df.
                write.
                partitionBy(partition_by).
                format(self.format).
                mode("append").
                options(**kwargs).
                save(str(self.dir))
            )
        else:
            logger.info(
                f"No new data to append to {self.name}. "
                "Nothing will be written."
            )

    def _dynamic_overwrite(
        self,
        df: ps.DataFrame,
        **kwargs
    ):
        """Overwrite partitions contained in the dataframe."""
        logger.info(
            f"Overwriting table partitions in {self.name}."
        )
        partition_by = self.partition_by
        if partition_by is None:
            partition_by = []
        (
            df.
            write.
            partitionBy(partition_by).
            format(self.format).
            mode("overwrite").
            options(**kwargs).
            save(str(self.dir))
        )

    def _check_for_null_unique_column_values(self, df: ps.DataFrame):
        """Remove a field from self.unique_column_set if all values are null."""
        for field_name in self.unique_column_set:
            if field_name in df.columns:
                if len(df.filter(df[field_name].isNotNull()).collect()) == 0:
                    logger.debug(
                        f"All {field_name} values are null. "
                        f"{field_name} will be removed as a partition column."
                    )
                    self.unique_column_set.remove(field_name)

    def _write_spark_df(
        self,
        df: ps.DataFrame,
        write_mode: TableWriteEnum = "append",
        num_partitions: int = None,
        **kwargs
    ):
        """Write spark dataframe to directory.

        Parameters
        ----------
        df : ps.DataFrame
            The spark dataframe to write.
        **kwargs
            Additional options to pass to the spark write method.
        """
        if self.ev.is_s3:
            logger.error("Writing to S3 is not supported.")
            raise ValueError("Writing to S3 is not supported.")

        logger.info(f"Writing files to {self.dir}.")

        if len(kwargs) == 0:
            kwargs = {
                "header": "true",
            }

        if df is not None:

            self._check_for_null_unique_column_values(df)

            if write_mode == "overwrite":
                self._dynamic_overwrite(df, **kwargs)
                self._load_table()
            elif write_mode == "append":
                self._append_without_duplicates(
                    df=df,
                    num_partitions=num_partitions,
                    **kwargs
                )
            elif write_mode == "upsert":
                self._upsert_without_duplicates(
                    df=df,
                    num_partitions=num_partitions,
                    **kwargs
                )
            else:
                raise ValueError(
                    f"Invalid write mode: {write_mode}. "
                    "Valid values are 'append', 'overwrite' and 'upsert'."
                )
            self._load_table()

    def _get_schema(self, type: str = "pyspark"):
        """Get the primary timeseries schema.

        Parameters
        ----------
        type : str, optional
            The type of schema to return.  Valid values are "pyspark" and
            "pandas".  The default is "pyspark".
        """
        if type == "pandas":
            return self.schema_func(type="pandas")

        return self.schema_func()

    def _validate(
        self,
        df: ps.DataFrame,
        strict: bool = True,
        add_missing_columns: bool = False
    ) -> ps.DataFrame:
        """Validate a DataFrame against the table schema.

        Parameters
        ----------
        df : ps.DataFrame
            The DataFrame to validate.
        strict : bool, optional
            If True, any extra columns will be dropped before validation.
            If False, will be validated as-is.
            The default is True.

        Returns
        -------
        validated_df : ps.DataFrame
            The validated DataFrame.
        """
        schema = self._get_schema()

        logger.info(f"Validating DataFrame with {schema.columns}.")

        schema_cols = schema.columns.keys()

        # Add missing columns
        if add_missing_columns:
            for col_name in schema_cols:
                if col_name not in df.columns:
                    df = df.withColumn(col_name, lit(None))

        if strict:
            df = df.select(*schema_cols)

        validated_df = schema.validate(df)

        if len(validated_df.pandera.errors) > 0:
            logger.error(f"Validation failed: {validated_df.pandera.errors}")
            raise ValueError(f"Validation failed: {validated_df.pandera.errors}")

        return validated_df

    def validate(self):
        """Validate the dataset table against the schema."""
        self._check_load_table()
        self._validate(self.df)

    def query(
        self,
        filters: Union[
            str, dict, FilterBaseModel,
            List[Union[str, dict, FilterBaseModel]]
        ] = None,
        order_by: Union[str, StrEnum, List[Union[str, StrEnum]]] = None
    ):
        """Run a query against the table with filters and order_by.

        In general a user will either use the query methods or the filter and
        order_by methods.  The query method is a convenience method that will
        apply filters and order_by in a single call.

        Parameters
        ----------
        filters : Union[
                str, dict, FilterBaseModel,
                List[Union[str, dict, FilterBaseModel]]
            ]
            The filters to apply to the query.  The filters can be an SQL string,
            dictionary, FilterBaseModel or a list of any of these. The filters
            will be applied in the order they are provided.

        order_by : Union[str, List[str], StrEnum, List[StrEnum]]
            The fields to order the query by.  The fields can be a string,
            StrEnum or a list of any of these.  The fields will be ordered in
            the order they are provided.

        Returns
        -------
        self : BaseTable or subclass of BaseTable

        Examples
        --------

        Filters as dictionaries:

        >>> ts_df = ev.primary_timeseries.query(
        >>>     filters=[
        >>>         {
        >>>             "column": "value_time",
        >>>             "operator": ">",
        >>>             "value": "2022-01-01",
        >>>         },
        >>>         {
        >>>             "column": "value_time",
        >>>             "operator": "<",
        >>>             "value": "2022-01-02",
        >>>         },
        >>>         {
        >>>             "column": "location_id",
        >>>             "operator": "=",
        >>>             "value": "gage-C",
        >>>         },
        >>>     ],
        >>>     order_by=["location_id", "value_time"]
        >>> ).to_pandas()

        Filters as SQL strings:

        >>> ts_df = ev.primary_timeseries.query(
        >>>     filters=[
        >>>         "value_time > '2022-01-01'",
        >>>         "value_time < '2022-01-02'",
        >>>         "location_id = 'gage-C'"
        >>>     ],
        >>>     order_by=["location_id", "value_time"]
        >>> ).to_pandas()

        Filters as FilterBaseModels:

        >>> from teehr.models.filters import TimeseriesFilter
        >>> from teehr.models.filters import FilterOperators
        >>>
        >>> fields = ev.primary_timeseries.field_enum()
        >>> ts_df = ev.primary_timeseries.query(
        >>>     filters=[
        >>>         TimeseriesFilter(
        >>>             column=fields.value_time,
        >>>             operator=FilterOperators.gt,
        >>>             value="2022-01-01",
        >>>         ),
        >>>         TimeseriesFilter(
        >>>             column=fields.value_time,
        >>>             operator=FilterOperators.lt,
        >>>             value="2022-01-02",
        >>>         ),
        >>>         TimeseriesFilter(
        >>>             column=fields.location_id,
        >>>             operator=FilterOperators.eq,
        >>>             value="gage-C",
        >>>         ),
        >>> ]).to_pandas()

        """
        logger.info("Performing the query.")
        self._check_load_table()
        if filters is not None:
            self.df = validate_and_apply_filters(
                sdf=self.df,
                filters=filters,
                filter_model=self.filter_model,
                fields_enum=self.field_enum(),
                dataframe_schema=self._get_schema("pandas"),
                validate=self.validate_filter_field_types
            )
        if order_by is not None:
            self.df = order_df(self.df, order_by)
        return self

    def filter(
        self,
        filters: Union[
            str, dict, FilterBaseModel,
            List[Union[str, dict, FilterBaseModel]]
        ]
    ):
        """Apply a filter.

        Parameters
        ----------
        filters : Union[
                str, dict, FilterBaseModel,
                List[Union[str, dict, FilterBaseModel]]
            ]
            The filters to apply to the query.  The filters can be an SQL string,
            dictionary, FilterBaseModel or a list of any of these.

        Returns
        -------
        self : BaseTable or subclass of BaseTable

        Examples
        --------

        Filters as dictionary:

        >>> ts_df = ev.primary_timeseries.filter(
        >>>     filters=[
        >>>         {
        >>>             "column": "value_time",
        >>>             "operator": ">",
        >>>             "value": "2022-01-01",
        >>>         },
        >>>         {
        >>>             "column": "value_time",
        >>>             "operator": "<",
        >>>             "value": "2022-01-02",
        >>>         },
        >>>         {
        >>>             "column": "location_id",
        >>>             "operator": "=",
        >>>             "value": "gage-C",
        >>>         },
        >>>     ]
        >>> ).to_pandas()

        Filters as string:

        >>> ts_df = ev.primary_timeseries.filter(
        >>>     filters=[
        >>>         "value_time > '2022-01-01'",
        >>>         "value_time < '2022-01-02'",
        >>>         "location_id = 'gage-C'"
        >>>     ]
        >>> ).to_pandas()

        Filters as FilterBaseModel:

        >>> from teehr.models.filters import TimeseriesFilter
        >>> from teehr.models.filters import FilterOperators
        >>>
        >>> fields = ev.primary_timeseries.field_enum()
        >>> ts_df = ev.primary_timeseries.filter(
        >>>     filters=[
        >>>         TimeseriesFilter(
        >>>             column=fields.value_time,
        >>>             operator=FilterOperators.gt,
        >>>             value="2022-01-01",
        >>>         ),
        >>>         TimeseriesFilter(
        >>>             column=fields.value_time,
        >>>             operator=FilterOperators.lt,
        >>>             value="2022-01-02",
        >>>         ),
        >>>         TimeseriesFilter(
        >>>             column=fields.location_id,
        >>>             operator=FilterOperators.eq,
        >>>             value="gage-C",
        >>>         ),
        >>> ]).to_pandas()

        """
        logger.info(f"Setting filter {filter}.")
        self._check_load_table()
        self.df = validate_and_apply_filters(
            sdf=self.df,
            filters=filters,
            filter_model=self.filter_model,
            fields_enum=self.field_enum(),
            dataframe_schema=self._get_schema("pandas"),
        )
        return self

    def order_by(
        self,
        fields: Union[str, StrEnum, List[Union[str, StrEnum]]]
    ):
        """Apply an order_by.

        Parameters
        ----------
        fields : Union[str, StrEnum, List[Union[str, StrEnum]]]
            The fields to order the query by.  The fields can be a string,
            StrEnum or a list of any of these.  The fields will be ordered in
            the order they are provided.

        Returns
        -------
        self : BaseTable or subclass of BaseTable

        Examples
        --------

        Order by string:

        >>> ts_df = ev.primary_timeseries.order_by("value_time").to_df()

        Order by StrEnum:

        >>> from teehr.querying.field_enums import TimeseriesFields
        >>> ts_df = ev.primary_timeseries.order_by(
        >>>     TimeseriesFields.value_time
        >>> ).to_pandas()

        """
        logger.info(f"Setting order_by {fields}.")
        self._check_load_table()
        self.df = order_df(self.df, fields)
        return self

    def fields(self) -> List[str]:
        """Return table columns as a list."""
        self._check_load_table()
        return self.df.columns

    def distinct_values(self, column: str) -> List[str]:
        """Return distinct values for a column."""
        self._check_load_table()
        return self.df.select(column).distinct().rdd.flatMap(lambda x: x).collect()

    def field_enum(self) -> StrEnum:
        """Get the fields enum."""
        raise NotImplementedError("field_enum method must be implemented.")

    def to_pandas(self):
        """Return Pandas DataFrame."""
        self._check_load_table()
        df = self.df.toPandas()
        df.attrs['table_type'] = self.name
        df.attrs['fields'] = self.fields()
        return df

    def to_geopandas(self):
        """Return GeoPandas DataFrame."""
        raise NotImplementedError("to_geopandas method must be implemented.")

    def to_sdf(self):
        """Return PySpark DataFrame.

        The PySpark DataFrame can be further processed using PySpark. Note,
        PySpark DataFrames are lazy and will not be executed until an action
        is called.  For example, calling `show()`, `collect()` or toPandas().
        This can be useful for further processing or analysis, for example,

        >>> ts_sdf = ev.primary_timeseries.query(
        >>>     filters=[
        >>>         "value_time > '2022-01-01'",
        >>>         "value_time < '2022-01-02'",
        >>>         "location_id = 'gage-C'"
        >>>     ]
        >>> ).to_sdf()
        >>> ts_df = (
        >>>     ts_sdf.select("value_time", "location_id", "value")
        >>>    .orderBy("value").toPandas()
        >>> )
        >>> ts_df.head()
        """
        self._check_load_table()
        return self.df




import abc
from pydantic import BaseModel as PydanticBaseModel, ConfigDict


class CalculatedFieldABC(abc.ABC):
    @abc.abstractmethod
    def apply_to(self):
        pass


class CalculatedFieldBaseModel(PydanticBaseModel):
    """Calculated field base model configuration."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra='forbid'  # raise an error if extra fields are passed
    )